Package: survdnn
Title: Deep Neural Networks for Survival Analysis with R 'torch'
Version: 0.7.6
Authors@R: 
    person(given = "Imad", 
           family = "EL BADISY", 
           email = "elbadisyimad@gmail.com", 
           role = c("aut", "cre"))
Description: Provides deep learning models for right-censored survival data using the 'torch' backend. 
    Supports multiple loss functions, including Cox partial likelihood, L2-penalized Cox, time-dependent Cox, 
    and accelerated failure time (AFT) loss. Offers a formula-based interface, built-in support for cross-validation, 
    hyperparameter tuning, survival curve plotting, and evaluation metrics such as the C-index, Brier score, 
    and integrated Brier score. For methodological details, see Kvamme et al. (2019) <https://www.jmlr.org/papers/v20/18-424.html>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: torch, survival, stats, utils, tibble, dplyr, purrr, tidyr,
        ggplot2, methods, rsample, cli, glue
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
RoxygenNote: 7.3.3
Config/testthat/edition: 3
BugReports: https://github.com/ielbadisy/survdnn/issues
URL: https://github.com/ielbadisy/survdnn
NeedsCompilation: no
Packaged: 2026-04-28 17:59:32 UTC; imad-el-badisy
Author: Imad EL BADISY [aut, cre]
Maintainer: Imad EL BADISY <elbadisyimad@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-29 06:40:13 UTC
Built: R 4.5.2; ; 2026-04-29 10:49:28 UTC; unix

Package: survdnn
Title: Deep Neural Networks for Survival Analysis with R 'torch'
Version: 0.7.6
Authors@R: 
    person(given = "Imad", 
           family = "EL BADISY", 
           email = "elbadisyimad@gmail.com", 
           role = c("aut", "cre"))
Description: Provides deep learning models for right-censored survival data using the 'torch' backend. 
    Supports multiple loss functions, including Cox partial likelihood, L2-penalized Cox, time-dependent Cox, 
    and accelerated failure time (AFT) loss. Offers a formula-based interface, built-in support for cross-validation, 
    hyperparameter tuning, survival curve plotting, and evaluation metrics such as the C-index, Brier score, 
    and integrated Brier score. For methodological details, see Kvamme et al. (2019) <https://www.jmlr.org/papers/v20/18-424.html>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: torch, survival, stats, utils, tibble, dplyr, purrr, tidyr,
        ggplot2, methods, rsample, cli, glue
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
RoxygenNote: 7.3.3
Config/testthat/edition: 3
BugReports: https://github.com/ielbadisy/survdnn/issues
URL: https://github.com/ielbadisy/survdnn
NeedsCompilation: no
Packaged: 2026-04-28 17:59:32 UTC; imad-el-badisy
Author: Imad EL BADISY [aut, cre]
Maintainer: Imad EL BADISY <elbadisyimad@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-29 06:40:13 UTC
Built: R 4.6.0; ; 2026-04-29 11:10:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,survdnn)
S3method(predict,survdnn)
S3method(print,survdnn)
S3method(summary,survdnn)
export(aft_loss)
export(brier)
export(build_dnn)
export(callback_early_stopping)
export(cindex_survmat)
export(cox_l2_loss)
export(cox_loss)
export(coxtime_loss)
export(cv_survdnn)
export(evaluate_survdnn)
export(gridsearch_survdnn)
export(ibs_survmat)
export(plot_loss)
export(summarize_cv_survdnn)
export(summarize_tune_survdnn)
export(survdnn)
export(tune_survdnn)
import(torch)
importFrom(dplyr,all_of)
importFrom(ggplot2,aes)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,theme_minimal)
importFrom(methods,is)
importFrom(stats,approx)
importFrom(stats,delete.response)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,sd)
importFrom(stats,terms)
importFrom(survival,Surv)
importFrom(survival,survfit)
importFrom(utils,tail)

# survdnn <img src="https://raw.githubusercontent.com/ielbadisy/survdnn/main/inst/logo.png" align="right" height="140"/>

> Deep Neural Networks for Survival Analysis using [R
> torch](https://torch.mlverse.org/)

[![License:
MIT](https://img.shields.io/badge/License-MIT-blue.svg)](LICENSE)  
[![R-CMD-check](https://github.com/ielbadisy/survdnn/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ielbadisy/survdnn/actions/workflows/R-CMD-check.yaml)

## About

`survdnn` implements neural network-based models for right-censored
survival analysis using the native `torch` backend in R. It supports
multiple loss functions including Cox partial likelihood, L2-penalized
Cox, Accelerated Failure Time (AFT) objectives, as well as
time-dependent extension such as Cox-Time. The package provides a
formula interface, supports model evaluation using time-dependent
metrics (C-index, Brier score, IBS), cross-validation, and
hyperparameter tuning.

## Review status

A methodological paper describing the design, implementation, and
evaluation of `survdnn` is currently under review at *The R Journal*.

## Main features

- Formula interface for `Surv() ~ .` models

- Modular neural architectures: configurable layers, activations,
  optimizers, and losses

- Built-in survival loss functions:

  - `"cox"`: Cox partial likelihood
  - `"cox_l2"`: penalized Cox
  - `"aft"`: Accelerated Failure Time
  - `"coxtime"`: deep time-dependent Cox

- Evaluation: C-index, Brier score, IBS

- Model selection with `cv_survdnn()` and `tune_survdnn()`

- Prediction of survival curves via `predict()` and `plot()`

## Installation

``` r
# Install from CRAN
install.packages("survdnn")


# Install from GitHub
install.packages("remotes")
remotes::install_github("ielbadisy/survdnn")

# Or clone and install locally
git clone https://github.com/ielbadisy/survdnn.git
setwd("survdnn")
devtools::install()
```

## Quick example

``` r
library(survdnn)
library(survival, quietly = TRUE)
library(ggplot2)

veteran <- survival::veteran

mod <- survdnn(
  Surv(time, status) ~ age + karno + celltype,
  data = veteran,
  hidden = c(32, 16),
  epochs = 300,
  loss = "cox",
  verbose = TRUE
  )

summary(mod)
```

    ## 
    ## Formula:
    ##   Surv(time, status) ~ age + karno + celltype
    ## <environment: 0x6011c835fbe0>
    ## 
    ## Model architecture:
    ##   Hidden layers:  32 : 16 
    ##   Activation:  relu 
    ##   Dropout:  0.3 
    ##   Batch norm:  TRUE 
    ##   Final loss:  3.841008 
    ## 
    ## Training summary:
    ##   Epochs:  300 
    ##   Learning rate:  1e-04 
    ##   Loss function:  cox 
    ##   Optimizer:  adam 
    ##   Device:  cpu 
    ##   CPU threads:  default 
    ##   NA action:  omit 
    ## 
    ## Data summary:
    ##   Observations (used/total): 137 / 137
    ##   Predictors (5): age, karno, celltypesmallcell, celltypeadeno, celltypelarge
    ##   Time range: [ 1, 999 ]
    ##   Events / censored:  128  /  9 
    ##   Event rate:  93.4% 
    ##   Predictors standardized:  yes

``` r
plot(mod, group_by = "celltype", times = 1:300)
```

## Loss Functions

- Cox partial likelihood

``` r
mod1 <- survdnn(
  Surv(time, status) ~ age + karno,
  data = veteran,
  loss = "cox",
  epochs = 300
  )
```

    ## [survdnn::fit] start: n=137 p=2 loss=cox optimizer=adam epochs=300 device=cpu

    ## [survdnn::fit] epoch 50/300 loss=3.989828

    ## [survdnn::fit] epoch 100/300 loss=3.942912

    ## [survdnn::fit] epoch 150/300 loss=3.890888

    ## [survdnn::fit] epoch 200/300 loss=3.869117

    ## [survdnn::fit] epoch 250/300 loss=3.866454

    ## [survdnn::fit] epoch 300/300 loss=3.870147

    ## [survdnn::fit] done: epochs_run=300 final_loss=3.870147

- Accelerated Failure Time

``` r
mod2 <- survdnn(
  Surv(time, status) ~ age + karno,
  data = veteran,
  loss = "aft",
  epochs = 300
  )
```

    ## [survdnn::fit] start: n=137 p=2 loss=aft optimizer=adam epochs=300 device=cpu

    ## [survdnn::fit] epoch 50/300 loss=4.628552

    ## [survdnn::fit] epoch 100/300 loss=4.641005

    ## [survdnn::fit] epoch 150/300 loss=4.563893

    ## [survdnn::fit] epoch 200/300 loss=4.507134

    ## [survdnn::fit] epoch 250/300 loss=4.501204

    ## [survdnn::fit] epoch 300/300 loss=4.494009

    ## [survdnn::fit] done: epochs_run=300 final_loss=4.494009

- Coxtime

``` r
mod3 <- survdnn(
  Surv(time, status) ~ age + karno,
  data = veteran,
  loss = "coxtime",
  epochs = 300
  )
```

    ## [survdnn::fit] start: n=137 p=2 loss=coxtime optimizer=adam epochs=300 device=cpu

    ## [survdnn::fit] epoch 50/300 loss=3.946552

    ## [survdnn::fit] epoch 100/300 loss=3.955686

    ## [survdnn::fit] epoch 150/300 loss=3.944864

    ## [survdnn::fit] epoch 200/300 loss=3.930768

    ## [survdnn::fit] epoch 250/300 loss=3.838923

    ## [survdnn::fit] epoch 300/300 loss=3.889966

    ## [survdnn::fit] done: epochs_run=300 final_loss=3.889966

## Cross-validation

``` r
cv_results <- cv_survdnn(
  Surv(time, status) ~ age + karno + celltype,
  data = veteran,
  times = c(600),
  metrics = c("cindex", "ibs"),
  folds = 3,
  hidden = c(16, 8),
  loss = "cox",
  epochs = 300
  )

print(cv_results)
```

## Hyperparameter tuning

``` r
grid <- list(
  hidden     = list(c(16), c(32, 16)),
  lr         = c(1e-3),
  activation = c("relu"),
  epochs     = c(100, 300),
  loss       = c("cox", "aft", "coxtime")
  )

tune_res <- tune_survdnn(
  formula = Surv(time, status) ~ age + karno + celltype,
  data = veteran,
  times = c(90, 300),
  metrics = "cindex",
  param_grid = grid,
  folds = 3,
  refit = FALSE,
  return = "summary"
  )

print(tune_res)
```

## Tuning and refitting the best Model

`tune_survdnn()` can be used also to automatically refit the
best-performing model on the full dataset. This behavior is controlled
by the `refit` and `return` arguments. For example:

``` r
best_model <- tune_survdnn(
  formula = Surv(time, status) ~ age + karno + celltype,
  data = veteran,
  times = c(90, 300),
  metrics = "cindex",
  param_grid = grid,
  folds = 3,
  refit = TRUE,
  return = "best_model"
  )
```

In this mode, cross-validation is used to select the optimal
hyperparameter configuration, after which the selected model is refitted
on the full dataset. The function then returns a fitted object of class
`"survdnn"`.

The resulting model can be used directly for prediction visualization,
and evaluation:

``` r
summary(best_model)

plot(best_model, times = 1:300)

predict(best_model, veteran, type = "risk", times = 180)
```

This makes `tune_survdnn()` suitable for end-to-end workflows, combining
model selection and final model fitting.

## Plot survival curves

``` r
plot(mod1, group_by = "celltype", times = 1:300)
```

![](README_files/figure-gfm/unnamed-chunk-11-1.png)<!-- -->

``` r
plot(mod1, group_by = "celltype", times = 1:300, plot_mean_only = TRUE)
```

![](README_files/figure-gfm/unnamed-chunk-12-1.png)<!-- -->

## Documentation

``` r
help(package = "survdnn")
?survdnn
?tune_survdnn
?cv_survdnn
?plot.survdnn
```

## Testing

``` r
# run all tests
devtools::test()
```

## Note on reproducibility

By default, `{torch}` initializes model weights and shuffles minibatches
using random draws, so results may differ across runs. Unlike
`set.seed()`, which only controls R’s random number generator, `{torch}`
relies on its own RNG implemented in C++ (and CUDA when using GPUs).

To ensure reproducibility, random seeds must therefore be set at the
Torch level as well.

`survdnn` provides built-in control of randomness to guarantee
reproducible results across runs. The main fitting function,
`survdnn()`, exposes a dedicated `.seed` argument:

``` r
mod <- survdnn(
  Surv(time, status) ~ age + karno + celltype,
  data   = veteran,
  epochs = 300,
  .seed  = 123
)
```

When `.seed` is provided, `survdnn()` internally synchronizes both R and
Torch random number generators via `survdnn_set_seed()`, ensuring
reproducible:

- weight initialization

- dropout behavior

- minibatch ordering

- loss trajectories

If `.seed = NULL` (the default), randomness is left uncontrolled and
results may vary between runs.

For full reproducibility in cross-validation or hyperparameter tuning,
the same `.seed` mechanism is propagated internally by `cv_survdnn()`
and `tune_survdnn()`, ensuring consistent data splits, model
initialization, and optimization paths across repetitions.

## CPU and core usage

`survdnn` relies on the `{torch}` backend for numerical computation. The
number of CPU cores (threads) used during training, prediction, and
evaluation is controlled globally by Torch.

By default, Torch automatically configures its CPU thread pools based on
the available system resources, unless explicitly overridden by the user
using:

``` r
torch::torch_set_num_threads(4)
```

You can also set this directly from `survdnn` APIs with `.threads`:

``` r
mod <- survdnn(
  Surv(time, status) ~ age + karno + celltype,
  data = survival::veteran,
  .threads = 4
)
```

    ## [survdnn::fit] start: n=137 p=5 loss=cox optimizer=adam epochs=300 device=cpu

    ## [survdnn::fit] cpu_threads=4

    ## [survdnn::fit] epoch 50/300 loss=3.962863

    ## [survdnn::fit] epoch 100/300 loss=3.952089

    ## [survdnn::fit] epoch 150/300 loss=3.896662

    ## [survdnn::fit] epoch 200/300 loss=3.834132

    ## [survdnn::fit] epoch 250/300 loss=3.866578

    ## [survdnn::fit] epoch 300/300 loss=3.821199

    ## [survdnn::fit] done: epochs_run=300 final_loss=3.821199

The same `.threads` argument is available in `cv_survdnn()` and
`tune_survdnn()`.

This setting affects:

- model training

- prediction

- evaluation metrics

- cross-validation and hyperparameter tuning

GPU acceleration can be enabled by setting `.device = "cuda"` when
calling `survdnn()` (`cv_survdnn()` and `tune_survdnn()` too).

## Availability

The `survdnn` R package is available on
[CRAN](https://CRAN.R-project.org/package=survdnn) or
[github](https://github.com/ielbadisy/survdnn)

## Contributions

Contributions, issues, and feature requests are welcome!

Open an [issue](https://github.com/ielbadisy/survdnn/issues) or submit a
pull request.

## License

MIT License © 2025 Imad EL BADISY

## handles silent setup
.onLoad <- function(libname, pkgname) {
  op <- options()
  op.survdnn <- list(survdnn.default_epochs = 100)
  toset <- !(names(op.survdnn) %in% names(op))
  if (any(toset)) options(op.survdnn[toset])

  utils::globalVariables(c(
    "fold", "metric", "value", "id", "time", "surv", "group", "mean_surv",
    "n", "se", "hidden", "lr", "activation", "epochs", "loss_name", ".loss_fn"
  ))
  # never load or probe torch here (because CRAN/Windows may segfault)
}

## handles user-facing messaging
.onAttach <- function(libname, pkgname) {
  # friendly hint that doesn't load the namespace:
  torch_pkg_present <- nzchar(system.file(package = "torch"))

  if (interactive() && !torch_pkg_present) {
    packageStartupMessage(
      "Optional dependency 'torch' not found. ",
      "Install the R package 'torch' and then run torch::install_torch() ",
      "to use deep-learning features."
    )
  }
  invisible()
}


## set R + torch seeds safely
survdnn_set_seed <- function(.seed = NULL) {
  if (is.null(.seed)) return(invisible(NULL))

  # R RNG
  set.seed(.seed)

  # torch RNG (only if available + backend installed)
  if (requireNamespace("torch", quietly = TRUE)) {
    if (isTRUE(torch::torch_is_installed())) {
      torch::torch_manual_seed(.seed)
    }
  }

  invisible(NULL)
}




## internal utility to choose a torch device for survdnn
survdnn_get_device <- function(.device = c("auto", "cpu", "cuda")) {
  .device <- match.arg(.device)

  if (!requireNamespace("torch", quietly = TRUE)) {
    stop(
      "The 'torch' package is required to fit survdnn models.\n",
      "Please install it with: install.packages('torch') and then run torch::install_torch().",
      call. = FALSE
    )
  }

  if (!isTRUE(torch::torch_is_installed())) {
    stop(
      "The Torch backend is not installed.\n",
      "Please run: torch::install_torch().",
      call. = FALSE
    )
  }

  if (.device == "cpu") {
    return(torch::torch_device("cpu"))
  }

  if (.device == "cuda") {
    if (!torch::cuda_is_available()) {
      warning("CUDA was requested but is not available; falling back to CPU.")
      return(torch::torch_device("cpu"))
    }
    return(torch::torch_device("cuda"))
  }

  if (torch::cuda_is_available()) {
    torch::torch_device("cuda")
  } else {
    torch::torch_device("cpu")
  }
}

test_that("survdnn_get_device falls back to CPU when torch is missing", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  dev_cpu <- survdnn_get_device("cpu")
  expect_s3_class(dev_cpu, "torch_device")
  expect_equal(dev_cpu$type, "cpu")
})

test_that("survdnn_get_device('auto') chooses CPU or CUDA consistently", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  dev <- survdnn_get_device("auto")
  expect_s3_class(dev, "torch_device")
  # Either cpu or cuda are acceptable
  expect_true(dev$type %in% c("cpu", "cuda"))
})

test_that("survdnn respects .device = 'cpu'", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  veteran <- survival::veteran

  mod_cpu <- survdnn(
    Surv(time, status) ~ age + karno,
    data    = veteran,
    epochs  = 2,
    loss    = "cox",
    verbose = FALSE,
    .device = "cpu"
  )

  expect_equal(mod_cpu$device$type, "cpu")
})

#' Build a Deep Neural Network for Survival Analysis
#'
#' Constructs a multilayer perceptron (MLP) with optional batch normalization
#' and dropout. Used internally by [survdnn()] to define the model architecture.
#'
#' @param input_dim Integer. Number of input features.
#' @param hidden Integer vector. Sizes of the hidden layers (e.g., c(32, 16)).
#' @param activation Character. Name of the activation function to use in each layer.
#'   Supported options: `"relu"`, `"leaky_relu"`, `"tanh"`, `"sigmoid"`, `"gelu"`, `"elu"`, `"softplus"`.
#' @param output_dim Integer. Output layer dimension (default = 1).
#' @param dropout Numeric between 0 and 1. Dropout rate after each hidden layer
#'   (default = 0.3). Set to 0 to disable dropout.
#' @param batch_norm Logical; whether to add `nn_batch_norm1d()` after each
#'   hidden linear layer (default = TRUE).
#'
#' @return A `nn_sequential` object representing the network.
#' @keywords internal
#' @export
build_dnn <- function(
  input_dim,
  hidden,
  activation = "relu",
  output_dim = 1L,
  dropout = 0.3,
  batch_norm = TRUE
) {
  layers <- list()
  in_features <- input_dim

  act_fn <- switch(
    activation,
    relu       = torch::nn_relu,
    leaky_relu = torch::nn_leaky_relu,
    tanh       = torch::nn_tanh,
    sigmoid    = torch::nn_sigmoid,
    gelu       = torch::nn_gelu,
    elu        = torch::nn_elu,
    softplus   = torch::nn_softplus,
    stop("Unsupported activation function: ", activation)
  )

  for (h in hidden) {
    layers <- append(layers, list(torch::nn_linear(in_features, h)))

    if (isTRUE(batch_norm)) {
      layers <- append(layers, list(torch::nn_batch_norm1d(h)))
    }

    layers <- append(layers, list(act_fn()))

    if (!is.null(dropout) && dropout > 0) {
      layers <- append(layers, list(torch::nn_dropout(p = dropout)))
    }

    in_features <- h
  }

  layers <- append(layers, list(torch::nn_linear(in_features, output_dim)))
  torch::nn_sequential(!!!layers)
}


#' Fit a Deep Neural Network for Survival Analysis
#'
#' Trains a deep neural network (DNN) to model right-censored survival data
#' using one of the predefined loss functions: Cox, AFT, or Coxtime.
#'
#' @param formula A survival formula of the form `Surv(time, status) ~ predictors`.
#' @param data A data frame containing the variables in the model.
#' @param hidden Integer vector. Sizes of the hidden layers (default: c(32, 16)).
#' @param activation Character string specifying the activation function to use in each layer.
#'   Supported options: `"relu"`, `"leaky_relu"`, `"tanh"`, `"sigmoid"`, `"gelu"`, `"elu"`, `"softplus"`.
#' @param lr Learning rate for the optimizer (default: `1e-4`).
#' @param epochs Number of training epochs (default: 300).
#' @param loss Character name of the loss function to use. One of `"cox"`, `"cox_l2"`, `"aft"`, or `"coxtime"`.
#' @param optimizer Character string specifying the optimizer to use. One of
#'   `"adam"`, `"adamw"`, `"sgd"`, `"rmsprop"`, or `"adagrad"`. Defaults to `"adam"`.
#' @param optim_args Optional named list of additional arguments passed to the
#'   underlying torch optimizer (e.g., `list(weight_decay = 1e-4, momentum = 0.9)`).
#' @param verbose Logical; whether to print progress and periodic loss updates during fitting (default: TRUE).
#' @param dropout Numeric between 0 and 1. Dropout rate applied after each
#'   hidden layer (default = 0.3). Set to 0 to disable dropout.
#' @param batch_norm Logical; whether to add batch normalization after each
#'   hidden linear layer (default = TRUE).
#' @param callbacks Optional list of callback functions. Each callback should have
#'   signature `function(epoch, current)` and return TRUE if training should stop,
#'   FALSE otherwise. Used, for example, with [callback_early_stopping()].
#' @param .seed Optional integer. If provided, sets both R and torch random seeds for reproducible
#'   weight initialization, shuffling, and dropout.
#' @param .device Character string indicating the computation device.
#'   One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"` uses CUDA if available,
#'   otherwise falls back to CPU.
#' @param .threads Optional positive integer. If provided, sets the global
#'   Torch CPU thread count via `torch::torch_set_num_threads()`.
#' @param na_action Character. How to handle missing values in the model variables:
#'   `"omit"` drops incomplete rows (and reports how many were removed when `verbose=TRUE`);
#'   `"fail"` stops with an error if any missing values are present.
#'
#' @return An object of class `"survdnn"` containing:
#' \describe{
#'   \item{model}{Trained `nn_module` object.}
#'   \item{formula}{Original survival formula.}
#'   \item{data}{Training data used for fitting.}
#'   \item{xnames}{Predictor variable names.}
#'   \item{x_center}{Column means of predictors.}
#'   \item{x_scale}{Column standard deviations of predictors.}
#'   \item{loss_history}{Vector of loss values per epoch.}
#'   \item{final_loss}{Final training loss.}
#'   \item{loss}{Loss function name used ("cox", "aft", etc.).}
#'   \item{activation}{Activation function used.}
#'   \item{hidden}{Hidden layer sizes.}
#'   \item{lr}{Learning rate.}
#'   \item{epochs}{Number of training epochs.}
#'   \item{optimizer}{Optimizer name used.}
#'   \item{optim_args}{List of optimizer arguments used.}
#'   \item{device}{Torch device used for training (`torch_device`).}
#'   \item{threads}{CPU thread setting passed via `.threads`; `NULL` means Torch default/global setting.}
#'   \item{aft_log_sigma}{Learned global log(sigma) for `loss="aft"`; `NA_real_` otherwise.}
#'   \item{aft_loc}{AFT log-time location offset used for centering when `loss="aft"`; `NA_real_` otherwise.}
#'   \item{coxtime_time_center}{Mean used to scale time for CoxTime; `NA_real_` otherwise.}
#'   \item{coxtime_time_scale}{SD used to scale time for CoxTime; `NA_real_` otherwise.}
#' }
#' @export
survdnn <- function(
  formula,
  data,
  hidden = c(32L, 16L),
  activation = "relu",
  lr = 1e-4,
  epochs = 300L,
  loss = c("cox", "cox_l2", "aft", "coxtime"),
  optimizer = c("adam", "adamw", "sgd", "rmsprop", "adagrad"),
  optim_args = list(),
  verbose = TRUE,
  dropout = 0.3,
  batch_norm = TRUE,
  callbacks = NULL,
  .seed = NULL,
  .device = c("auto", "cpu", "cuda"),
  .threads = NULL,
  na_action = c("omit", "fail")
) {
  survdnn_set_seed(.seed)
  device <- survdnn_get_device(.device)
  survdnn_set_threads(.threads)

  loss      <- match.arg(loss)
  optimizer <- match.arg(optimizer)
  na_action <- match.arg(na_action)

  if (!is.list(optim_args)) {
    stop("`optim_args` must be a list (possibly empty).", call. = FALSE)
  }

  if (!is.null(callbacks)) {
    if (is.function(callbacks)) {
      callbacks <- list(callbacks)
    } else if (!is.list(callbacks) || !all(vapply(callbacks, is.function, logical(1)))) {
      stop("`callbacks` must be NULL, a function, or a list of functions.", call. = FALSE)
    }
  }

  stopifnot(inherits(formula, "formula"))
  stopifnot(is.data.frame(data))

  environment(formula) <- list2env(
    list(Surv = survival::Surv),
    parent = environment(formula)
  )

  # missing data handling
  n_before <- nrow(data)
  mf <- model.frame(
    formula,
    data = data,
    na.action = if (na_action == "omit") stats::na.omit else stats::na.fail
  )
  n_after <- nrow(mf)
  n_removed <- n_before - n_after
  if (n_removed > 0 && isTRUE(verbose) && na_action == "omit") {
    message(sprintf("[survdnn::fit] removed %d observations with missing values.", n_removed))
  }

  y        <- model.response(mf)
  x        <- model.matrix(attr(mf, "terms"), data = mf)[, -1, drop = FALSE]
  time     <- y[, "time"]
  status   <- y[, "status"]
  x_scaled <- scale(x)

  if (isTRUE(verbose)) {
    message(
      sprintf(
        "[survdnn::fit] start: n=%d p=%d loss=%s optimizer=%s epochs=%d device=%s",
        nrow(mf), ncol(x), loss, optimizer, epochs, as.character(device)
      )
    )
    if (!is.null(.threads)) {
      message(sprintf("[survdnn::fit] cpu_threads=%d", as.integer(.threads)))
    }
  }

  # AFT location offset for stability
  aft_loc <- NA_real_
  if (loss == "aft") {
    evt <- (status == 1)
    if (any(evt)) {
      aft_loc <- mean(log(pmax(time[evt], .Machine$double.eps)))
    } else {
      aft_loc <- mean(log(pmax(time, .Machine$double.eps)))
    }
    if (!is.finite(aft_loc)) aft_loc <- 0
  }

  # CoxTime time scaling
  coxtime_time_center <- NA_real_
  coxtime_time_scale  <- NA_real_
  time_scaled <- NULL

  if (loss == "coxtime") {
    ts <- scale(as.numeric(time))
    coxtime_time_center <- as.numeric(attr(ts, "scaled:center"))
    coxtime_time_scale  <- as.numeric(attr(ts, "scaled:scale"))
    if (!is.finite(coxtime_time_scale) || coxtime_time_scale <= 0) coxtime_time_scale <- 1
    time_scaled <- as.numeric(ts)
  }

  # x_tensor:
  # - only for coxtime: [time_scaled, x_scaled]  (time as fed to net)
  # - others : [x_scaled]
  x_tensor <- if (loss == "coxtime") {
    torch::torch_tensor(
      cbind(time_scaled, x_scaled),
      dtype  = torch::torch_float(),
      device = device
    )
  } else {
    torch::torch_tensor(
      x_scaled,
      dtype  = torch::torch_float(),
      device = device
    )
  }

  # y_tensor always uses RAW time for ordering/risk sets
  y_tensor <- torch::torch_tensor(
    cbind(time, status),
    dtype  = torch::torch_float(),
    device = device
  )

  # network
  net <- build_dnn(
    input_dim  = ncol(x_tensor),
    hidden     = hidden,
    activation = activation,
    output_dim = 1L,
    dropout    = dropout,
    batch_norm = batch_norm
  )
  net$to(device = device)

  # loss dispatcher + (optional) AFT extra params
  extra_params  <- NULL # list for AFT, NULL otherwise
  aft_log_sigma <- NA_real_     
  loss_fn <- NULL

  if (loss == "cox") {
    loss_fn <- function(net, x, y) cox_loss(net(x), y)
  } else if (loss == "cox_l2") {
    loss_fn <- function(net, x, y) cox_l2_loss(net(x), y, lambda = 1e-3)
  } else if (loss == "aft") {
    loc0 <- if (is.finite(aft_loc)) aft_loc else 0
    aft_bundle <- survdnn__aft_lognormal_nll_factory(device = device, aft_loc = loc0)
    extra_params <- aft_bundle$extra_params
    loss_fn <- function(net, x, y) aft_bundle$loss_fn(net, x, y)
  } else if (loss == "coxtime") {
    lf <- survdnn__coxtime_loss_factory(net)
    loss_fn <- function(net, x, y) lf(x, y)
  } else {
    stop("Unsupported loss: ", loss, call. = FALSE)
  }

  # optimizer params
  params <- net$parameters
  if (loss == "aft" && !is.null(extra_params) && !is.null(extra_params$log_sigma)) {
    params <- c(params, list(extra_params$log_sigma))
  }

  opt_args <- c(list(params = params, lr = lr), optim_args)

  if (is.null(optim_args$weight_decay) && optimizer %in% c("adam", "adamw")) {
    opt_args$weight_decay <- 1e-4
  }

  optimizer_obj <- switch(
    optimizer,
    adam    = do.call(torch::optim_adam,    opt_args),
    adamw   = do.call(torch::optim_adamw,   opt_args),
    sgd     = do.call(torch::optim_sgd,     opt_args),
    rmsprop = do.call(torch::optim_rmsprop, opt_args),
    adagrad = do.call(torch::optim_adagrad, opt_args),
    stop("Unsupported optimizer: ", optimizer)
  )

  # training loop
  loss_history   <- numeric(epochs)
  early_stopped  <- FALSE
  last_epoch_run <- epochs

  for (epoch in 1:epochs) {
    net$train()
    optimizer_obj$zero_grad()

    loss_val <- loss_fn(net, x_tensor, y_tensor)
    loss_val$backward()
    optimizer_obj$step()

    current_loss        <- loss_val$item()
    loss_history[epoch] <- current_loss
    last_epoch_run      <- epoch

    if (isTRUE(verbose) && (epoch %% 50 == 0 || epoch == epochs)) {
      message(sprintf("[survdnn::fit] epoch %d/%d loss=%.6f", epoch, epochs, current_loss))
    }

    if (!is.null(callbacks)) {
      for (cb in callbacks) {
        if (isTRUE(cb(epoch, current_loss))) {
          early_stopped <- TRUE
          break
        }
      }
      if (early_stopped) break
    }
  }

  if (early_stopped && last_epoch_run < epochs) {
    loss_history <- loss_history[seq_len(last_epoch_run)]
    if (isTRUE(verbose)) {
      message(sprintf("[survdnn::fit] early stopping at epoch %d/%d.", last_epoch_run, epochs))
    }
  }

  # store learned AFT log(sigma) robustly
  if (loss == "aft" && !is.null(extra_params) && !is.null(extra_params$log_sigma)) {
    aft_log_sigma <- as.numeric(extra_params$log_sigma$item())
    if (!is.finite(aft_log_sigma)) aft_log_sigma <- NA_real_
  } else {
    aft_log_sigma <- NA_real_
  }

  if (isTRUE(verbose)) {
    message(
      sprintf(
        "[survdnn::fit] done: epochs_run=%d final_loss=%.6f",
        last_epoch_run,
        tail(loss_history, 1)
      )
    )
  }

  structure(
    list(
      model               = net,
      formula             = formula,
      data                = data,
      xnames              = colnames(x),
      x_center            = attr(x_scaled, "scaled:center"),
      x_scale             = attr(x_scaled, "scaled:scale"),
      loss_history        = loss_history,
      final_loss          = tail(loss_history, 1),
      loss                = loss,
      activation          = activation,
      hidden              = hidden,
      lr                  = lr,
      epochs              = epochs,
      optimizer           = optimizer,
      optim_args          = optim_args,
      device              = device,
      threads             = .threads,
      dropout             = dropout,
      batch_norm          = batch_norm,
      na_action           = na_action,
      aft_log_sigma       = aft_log_sigma,
      aft_loc             = if (loss == "aft") aft_loc else NA_real_,
      coxtime_time_center = if (loss == "coxtime") coxtime_time_center else NA_real_,
      coxtime_time_scale  = if (loss == "coxtime") coxtime_time_scale  else NA_real_
    ),
    class = "survdnn"
  )
}

test_that("survdnn omits NAs and messages when verbose=TRUE", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())


  dat <- data.frame(
    time  = c(1, 2, 3, 4),
    status = c(1, 1, 0, 1),
    x1    = c(0.1, NA, 0.3, 0.4)
  )

  expect_message(
    survdnn(
      survival::Surv(time, status) ~ x1,
      data = dat,
      na_action = "omit",
      verbose = TRUE,
      epochs = 1
    ),
    "Removed 1 observations with missing values"
  )
})

test_that("survdnn fails on NAs when na_action='fail'", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())


  dat <- data.frame(
    time  = c(1, 2, 3, 4),
    status = c(1, 1, 0, 1),
    x1    = c(0.1, NA, 0.3, 0.4)
  )

  expect_error(
    survdnn(
      survival::Surv(time, status) ~ x1,
      data = dat,
      na_action = "fail",
      verbose = FALSE,
      epochs = 1
    )
  )
})

test_that("evaluate_survdnn respects na_action='fail' on newdata", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())


  train <- data.frame(
    time  = c(1, 2, 3, 4, 5, 6),
    status = c(1, 1, 0, 1, 0, 1),
    x1    = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
  )

  mod <- survdnn(
    survival::Surv(time, status) ~ x1,
    data = train,
    epochs = 2,
    verbose = FALSE,
    na_action = "fail"
  )

  newdat <- data.frame(
    time  = c(1, 2, 3),
    status = c(1, 0, 1),
    x1    = c(0.1, NA, 0.3)
  )

  expect_error(
    evaluate_survdnn(
      mod,
      metrics = "cindex",
      times = 2,
      newdata = newdat,
      na_action = "fail"
    )
  )
})



test_that("evaluate_survdnn omits incomplete rows when na_action='omit' (no crash)", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())


  train <- data.frame(
    time   = c(1, 2, 3, 4, 5, 6),
    status = c(1, 1, 0, 1, 0, 1),
    x1     = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
  )

  mod <- survdnn(
    survival::Surv(time, status) ~ x1,
    data = train,
    epochs = 2,
    verbose = FALSE,
    na_action = "fail"
  )

  newdat <- data.frame(
    time   = c(1, 2, 3),
    status = c(1, 0, 1),
    x1     = c(0.1, NA, 0.3)
  )

  res <- suppressWarnings(
    evaluate_survdnn(
      mod,
      metrics = c("brier", "ibs"),
      times = c(1, 2),
      newdata = newdat,
      na_action = "omit",
      verbose = TRUE
    )
  )

  expect_s3_class(res, "tbl_df")
  expect_true(all(c("metric", "value") %in% names(res)))
})

test_that("evaluate_survdnn returns expected metric values", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  times <- c(30, 90, 180)

  out <- evaluate_survdnn(mod, metrics = c("cindex", "ibs", "brier"), times = times)

  expect_s3_class(out, "data.frame")
  expect_true(all(c("metric", "value") %in% names(out)))
  expect_true(any(out$metric == "cindex"))
  expect_true(any(out$metric == "ibs"))
  expect_true(any(out$metric == "brier"))
})

test_that("evaluate_survdnn errors with unknown metrics or missing times", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  expect_error(evaluate_survdnn(mod, metrics = "abc", times = 100), "Unknown metric")
  expect_error(evaluate_survdnn(mod, metrics = "cindex"), "must provide `times`")
})

test_that("summary.survdnn prints key summary components", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  output <- capture.output(summary(mod))

  expect_true(any(grepl("Model architecture", output)))
  expect_true(any(grepl("Data summary", output)))
  expect_true(any(grepl("Event rate", output)))

})

test_that("callback_early_stopping stops on plateauing loss", {
  cb <- callback_early_stopping(patience = 2, mode = "min")

  # fake loss sequence: improves, then flat/worse
  out1 <- cb(epoch = 1, current = 1.0)  # best = 1.0
  out2 <- cb(epoch = 2, current = 0.9)  # best = 0.9
  out3 <- cb(epoch = 3, current = 0.91) # wait = 1
  out4 <- cb(epoch = 4, current = 0.92) # wait = 2 -> stop

  expect_false(out1)
  expect_false(out2)
  expect_false(out3)
  expect_true(out4)
})

test_that("survdnn uses early stopping callback on training loss", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  veteran <- survival::veteran

  # strong early stopping: patience small, epochs large (need to investigate this in a simuation study)
  cb <- callback_early_stopping(patience = 1, mode = "min")

  mod <- survdnn::survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data       = veteran,
    hidden     = c(8L, 4L),
    activation = "relu",
    epochs     = 50L,
    loss       = "cox",
    verbose    = FALSE,
    dropout    = 0.3,
    batch_norm = TRUE,
    callbacks  = cb,
    .seed      = 123,
    .device    = "cpu"
  )

  # we can't guarantee how early it stops, but in realistic settings
  # we expect fewer than the requested epochs when callback is active (more investigation is needed)
  expect_true(length(mod$loss_history) <= 50L)
  expect_true(length(mod$loss_history) >= 1L)
})

test_that("survdnn accepts multiple optimizers", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  veteran <- survival::veteran

  opts <- c("adam", "adamw", "sgd", "rmsprop", "adagrad")

  for (opt in opts) {
    mod <- survdnn(
      Surv(time, status) ~ age + karno + celltype,
      data       = veteran,
      hidden     = c(8L, 4L),
      activation = "relu",
      lr         = 1e-3,
      epochs     = 3L,
      loss       = "cox",
      optimizer  = opt,
      verbose    = FALSE,
      .device    = "cpu"
    )

    expect_s3_class(mod, "survdnn")
    expect_equal(mod$optimizer, opt)
    expect_true(length(mod$loss_history) >= 1L)
  }
})

test_that("optim_args is passed to optimizer", {
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  veteran <- survival::veteran

  mod <- survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data       = veteran,
    hidden     = c(8L, 4L),
    activation = "relu",
    lr         = 1e-3,
    epochs     = 3L,
    loss       = "cox",
    optimizer  = "sgd",
    optim_args = list(momentum = 0.9),
    verbose    = FALSE,
    .device    = "cpu"
  )

  expect_s3_class(mod, "survdnn")
  expect_equal(mod$optimizer, "sgd")
})

test_that("plot_loss returns a ggplot object", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  veteran <- survival::veteran

  mod <- survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data   = veteran,
    epochs = 5,
    loss   = "cox",
    verbose = FALSE,
    .device = "cpu"
  )

  p <- plot_loss(mod)
  expect_s3_class(p, "ggplot")
})

test_that("print.survdnn prints without error and returns invisibly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  expect_invisible(print(mod))
  expect_s3_class(mod, "survdnn")
})

test_that("predict.survdnn works for type = 'lp'", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 10, verbose = FALSE)

  lp <- predict(mod, newdata = data, type = "lp")

  expect_type(lp, "double")
  expect_length(lp, nrow(data))
  expect_false(anyNA(lp))
})

test_that("predict.survdnn works for type = 'survival'", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  times <- c(30, 90, 180)
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 10, verbose = FALSE)

  surv <- predict(mod, newdata = data, type = "survival", times = times)

  expect_s3_class(surv, "data.frame")
  expect_equal(ncol(surv), length(times))
  expect_equal(nrow(surv), nrow(data))
  expect_true(all(surv >= 0 & surv <= 1))
  expect_named(surv, paste0("t=", times))
})

test_that("predict.survdnn works for type = 'risk'", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 10, verbose = FALSE)

  risk <- predict(mod, newdata = data, type = "risk", times = 180)

  expect_type(risk, "double")
  expect_length(risk, nrow(data))
  expect_true(all(risk >= 0 & risk <= 1))
})

test_that("predict.survdnn handles invalid or missing times properly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 10, verbose = FALSE)

  expect_error(predict(mod, newdata = data, type = "survival"),
               "`times` must be specified")

  
  expect_error(predict(mod, newdata = data, type = "risk", times = c(30, 60)),
                "For type = 'risk', `times` must be a single( numeric)? value\\.")
              
})

test_that("plot.survdnn returns a ggplot object (default)", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, hidden = c(8), epochs = 5, verbose = FALSE)

  p <- plot(mod)

  expect_s3_class(p, "ggplot")
})

test_that("plot.survdnn supports group_by and plot_mean_only", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, hidden = c(8), epochs = 5, verbose = FALSE)

  # grouped + mean only
  p1 <- plot(mod, group_by = "celltype", plot_mean_only = TRUE, times = 1:100)
  expect_s3_class(p1, "ggplot")

  # grouped + individual + mean
  p2 <- plot(mod, group_by = "celltype", plot_mean_only = FALSE, add_mean = TRUE, times = 1:100)
  expect_s3_class(p2, "ggplot")
})

test_that("plot.survdnn fails with wrong inputs", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  expect_error(survdnn:::plot.survdnn("not a model"), "inherits")
})

test_that("survdnn() fits a model and returns correct structure", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  set.seed(123)
  n <- 100
  df <- data.frame(
    time   = rexp(n, rate = 0.1),
    status = rbinom(n, 1, 0.7),
    x1     = rnorm(n),
    x2     = rbinom(n, 1, 0.5)
  )

  mod <- survdnn(
    Surv(time, status) ~ x1 + x2,
    data = df,
    hidden = c(8, 4),
    activation = "relu",
    epochs = 10,
    lr = 1e-3,
    loss = "cox",
    verbose = FALSE
  )

  expect_s3_class(mod, "survdnn")
  expect_true(inherits(mod$model, "nn_module"))
  expect_equal(deparse(mod$formula), deparse(Surv(time, status) ~ x1 + x2))
  expect_type(mod$final_loss, "double")
  expect_length(mod$loss_history, 10)

  expect_named(
    mod,
    c(
      "activation",
      "aft_loc",
      "aft_log_sigma",
      "batch_norm",
      "coxtime_time_center",
      "coxtime_time_scale",
      "data",
      "device",
      "dropout",
      "epochs",
      "final_loss",
      "formula",
      "hidden",
      "loss",
      "loss_history",
      "lr",
      "model",
      "na_action",
      "optim_args",
      "optimizer",
      "threads",
      "x_center",
      "x_scale",
      "xnames"
    ),
    ignore.order = TRUE
  )

  # sanity: for loss="cox", AFT/CoxTime metadata should be NA
  expect_true(is.na(mod$aft_log_sigma))
  expect_true(is.na(mod$aft_loc))
  expect_true(is.na(mod$coxtime_time_center))
  expect_true(is.na(mod$coxtime_time_scale))
})

test_that("survdnn() validates .threads", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  set.seed(123)
  df <- data.frame(
    time   = rexp(20, rate = 0.1),
    status = rbinom(20, 1, 0.7),
    x1     = rnorm(20)
  )

  expect_error(
    survdnn(
      Surv(time, status) ~ x1,
      data = df,
      epochs = 1,
      verbose = FALSE,
      .threads = 0
    ),
    "`.threads` must be a single positive integer or NULL."
  )
})


test_that("survdnn() is reproducible given .seed", {
  skip_on_cran()
  skip_if_not_installed("torch")
  skip_if_not(torch::torch_is_installed())

  set.seed(123)
  n <- 80
  df <- data.frame(
    time   = rexp(n, rate = 0.1),
    status = rbinom(n, 1, 0.7),
    x1     = rnorm(n),
    x2     = rbinom(n, 1, 0.5)
  )

  mod1 <- survdnn(
    Surv(time, status) ~ x1 + x2,
    data   = df,
    hidden = c(8),
    activation = "relu",
    lr     = 1e-3,
    epochs = 10,
    loss   = "cox",
    verbose = FALSE,
    .seed   = 999
  )

  mod2 <- survdnn(
    Surv(time, status) ~ x1 + x2,
    data   = df,
    hidden = c(8),
    activation = "relu",
    lr     = 1e-3,
    epochs = 10,
    loss   = "cox",
    verbose = FALSE,
    .seed   = 999
  )

  expect_equal(mod1$loss_history, mod2$loss_history)
})

#' Plot survdnn Survival Curves using ggplot2
#'
#' Visualizes survival curves predicted by a fitted `survdnn` model.
#' Curves can be grouped by a categorical variable in `newdata` and
#' optionally display only the group-wise means or overlay them.
#'
#' @param x A fitted `survdnn` model object.
#' @param newdata Optional data frame for prediction (defaults to training data).
#' @param times A numeric vector of time points at which to compute survival probabilities.
#' @param group_by Optional name of a column in `newdata` used to color and group curves.
#' @param plot_mean_only Logical; if `TRUE`, plots only the mean survival curve per group.
#' @param add_mean Logical; if `TRUE`, adds mean curves to the individual lines.
#' @param alpha Alpha transparency for individual curves (ignored if `plot_mean_only = TRUE`).
#' @param mean_lwd Line width for mean survival curves.
#' @param mean_lty Line type for mean survival curves.
#' @param ... Reserved for future use.
#'
#' @return A `ggplot` object.
#' @export
#'
#' @examples
#' \donttest{
#' library(survival)
#' data(veteran)
#' set.seed(42)
#' \donttest{
#' mod <- survdnn(Surv(time, status) ~ age + karno + celltype, data = veteran,
#'                hidden = c(16, 8), epochs = 100, verbose = FALSE)
#' plot(mod, group_by = "celltype", times = 1:300)
#' }
#' }
plot.survdnn <- function(x, newdata = NULL, times = 1:365,
                         group_by = NULL,
                         plot_mean_only = FALSE,
                         add_mean = TRUE,
                         alpha = 0.3,
                         mean_lwd = 1.3,
                         mean_lty = 1,
                         ...) {
  stopifnot(inherits(x, "survdnn"))
  if (is.null(newdata)) newdata <- x$data

  # compute survival probabilities
  survmat <- predict(x, newdata = newdata, times = times, type = "survival")
  df_surv <- as.data.frame(survmat)
  df_surv$id <- seq_len(nrow(df_surv))

  # reshape to long format
  df_long <- tidyr::pivot_longer(
    df_surv,
    cols = -id,
    names_to = "time_label",
    values_to = "surv"
  )

  # clean up time labels
  df_long$time <- as.numeric(gsub("t=", "", df_long$time_label))

  # group handling
  if (!is.null(group_by)) {
    if (!group_by %in% names(newdata)) {
      stop("Column '", group_by, "' not found in newdata.")
    }
    df_long$group <- rep(newdata[[group_by]], each = length(times))
  } else {
    df_long$group <- "all"
  }

  # base ggplot
  p <- ggplot(df_long, aes(x = time, y = surv, group = id, color = group))

  # plot individual curves
  if (!plot_mean_only) {
    p <- p + geom_line(alpha = alpha, linewidth = 0.7)
  }

  # plot mean curves
  if (add_mean || plot_mean_only) {
    df_mean <- dplyr::group_by(df_long, group, time) |>
      dplyr::summarise(mean_surv = mean(surv, na.rm = TRUE), .groups = "drop")

    p <- p + geom_line(data = df_mean,
                       mapping = aes(x = time, y = mean_surv, color = group),
                       linewidth = mean_lwd,
                       linetype = mean_lty,
                       inherit.aes = FALSE)
  }

  # finalize
  p <- p +
    theme_minimal() +
    labs(
      title = "Predicted Survival Curves",
      x = "Time",
      y = "Survival Probability",
      color = if (!is.null(group_by)) group_by else NULL
    )

  return(p)
}

#' Grid Search for survdnn Hyperparameters
#'
#' Performs grid search over user-specified hyperparameters and evaluates performance on a validation set.
#'
#' @param formula A survival formula (e.g., `Surv(time, status) ~ .`)
#' @param train Training dataset
#' @param valid Validation dataset
#' @param times Evaluation time points (numeric vector)
#' @param metrics Evaluation metrics (character vector): any of "cindex" and "ibs".
#' @param param_grid A named list of hyperparameters to search over. Currently
#'   supported entries are \code{hidden}, \code{lr}, \code{activation},
#'   \code{epochs}, and \code{loss}.
#' @param .seed Optional random seed for reproducibility
#' @param .device Character string indicating the computation device used when
#'   fitting all models in the grid search. One of \code{"auto"}, \code{"cpu"},
#'   or \code{"cuda"}. This is a runtime setting and is not part of the
#'   hyperparameter grid.
#' 
#' @return A tibble with configurations and their validation metrics
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   set.seed(123)
#'
#'   # Simulate small dataset
#'   n <- 60
#'   x1 <- rnorm(n); x2 <- rbinom(n, 1, 0.5)
#'   time <- rexp(n, rate = 0.1)
#'   status <- rbinom(n, 1, 0.7)
#'   df <- data.frame(time, status, x1, x2)
#'
#'   # Split into training and validation
#'   idx <- sample(seq_len(n), 0.7 * n)
#'   train <- df[idx, ]
#'   valid <- df[-idx, ]
#'
#'   # Define formula and param grid
#'   formula <- survival::Surv(time, status) ~ x1 + x2
#'   param_grid <- list(
#'     hidden     = list(c(4)),
#'     lr         = c(1e-3),
#'     activation = c("relu"),
#'     epochs     = c(1),
#'     loss       = c("cox")
#'   )
#'
#'   # Run grid search
#'   results <- gridsearch_survdnn(
#'     formula = formula,
#'     train   = train,
#'     valid   = valid,
#'     times   = c(10, 20),
#'     metrics = c("cindex", "ibs"),
#'     param_grid = param_grid
#'   )
#'
#'   # View summary
#'   dplyr::group_by(results, hidden, lr, activation, epochs, loss, metric) |>
#'     dplyr::summarise(mean = mean(value, na.rm = TRUE), .groups = "drop")
#' }
#' }

gridsearch_survdnn <- function(formula, train, valid, times,
  metrics = c("cindex", "ibs"),
  param_grid,
  .seed = 42,
  .device = c("auto", "cpu", "cuda")) {
  .device <- match.arg(.device)
  
  if (!is.null(.seed)) survdnn_set_seed(.seed)
  
  req <- c("hidden","lr","activation","epochs","loss")
  miss <- setdiff(req, names(param_grid))
  if (length(miss)) stop("param_grid is missing: ", paste(miss, collapse=", "), call. = FALSE)

  
  param_df <- tidyr::crossing(!!!param_grid)
  
  
  results <- purrr::pmap_dfr(param_df, function(hidden, lr, activation, epochs, loss) {
    
    message(glue::glue("[survdnn] Training: loss={loss}, activation={activation}, hidden={toString(hidden)}"))

# re-seed inside each config to make results fully reproducible
survdnn_set_seed(.seed)

mod <- survdnn(
  formula    = formula,
  data       = train,
  hidden     = hidden,
  lr         = lr,
  activation = activation,
  epochs     = epochs,
  loss       = loss,
  verbose    = FALSE,
  .seed      = .seed,
  .device    = .device
)

eval_tbl <- evaluate_survdnn(
  model    = mod,
  newdata  = valid,
  metrics  = metrics,
  times    = times
)

config <- tibble::tibble(
  hidden     = list(hidden),
  lr         = lr,
  activation = activation,
  epochs     = epochs,
  loss       = loss
)

dplyr::bind_cols(config[rep(1, nrow(eval_tbl)), ], eval_tbl)
  })

return(results)
}

#' Concordance Index from a Survival Probability Matrix
#'
#' Computes the time-dependent concordance index (C-index) from a predicted survival matrix
#' at a fixed time point. The risk is computed as `1 - S(t_star)`.
#'
#' @param object A `Surv` object representing the observed survival data.
#' @param predicted A data frame or matrix of predicted survival probabilities.
#'   Each column corresponds to a time point (e.g., `t=90`, `t=180`).
#' @param t_star A numeric time point corresponding to one of the columns in `predicted`.
#'   If `NULL`, the last column is used.
#'
#' @return A single numeric value representing the C-index.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   veteran <- survival::veteran
#'   mod <- survdnn(survival::Surv(time, status) ~
#'   age + karno + celltype, data = veteran, epochs = 50, verbose = FALSE)
#'   pred <- predict(mod, newdata = veteran, type = "survival", times = c(30, 90, 180))
#'   y <- model.response(model.frame(mod$formula, veteran))
#'   cindex_survmat(y, pred, t_star = 180)
#' }
#' }

cindex_survmat <- function(object, predicted, t_star = NULL) {
  if (!inherits(object, "Surv")) stop("object must be a survival object (from Surv())")

  time <- object[, 1]
  status <- object[, 2]

  if (!is.null(t_star)) {
    t_name <- paste0("t=", t_star)
    if (!(t_name %in% colnames(predicted))) {
      stop("t_star = ", t_star, " not found in predicted survival matrix.")
    }
    surv_prob <- predicted[[t_name]]
  } else {
    surv_prob <- predicted[[ncol(predicted)]]
  }

  risk_score <- 1 - surv_prob
  n <- length(time)

  concord <- 0
  par_concord <- 0
  permissible <- 0

  for (i in 1:(n - 1)) {
    for (j in (i + 1):n) {
      if ((time[i] < time[j] && status[i] == 0) ||
          (time[j] < time[i] && status[j] == 0)) next
      if (time[i] == time[j] && status[i] == 0 && status[j] == 0) next

      permissible <- permissible + 1

      if (time[i] != time[j]) {
        if ((time[i] < time[j] && risk_score[i] > risk_score[j]) ||
            (time[j] < time[i] && risk_score[j] > risk_score[i])) {
          concord <- concord + 1
        } else if (risk_score[i] == risk_score[j]) {
          par_concord <- par_concord + 0.5
        }
      } else {
        if (status[i] + status[j] > 0) {
          if (risk_score[i] == risk_score[j]) {
            concord <- concord + 1
          } else {
            par_concord <- par_concord + 0.5
          }
        }
      }
    }
  }

  c_index <- (concord + par_concord) / permissible
  names(c_index) <- "c-index"
  return(round(c_index, 6))
}


#' Brier Score for Right-Censored Survival Data at a Fixed Time
#'
#' Computes the Brier score at a fixed time point using inverse probability of censoring weights (IPCW).
#'
#' @param object A `Surv` object with observed time and status.
#' @param pre_sp A numeric vector of predicted survival probabilities at `t_star`.
#' @param t_star The evaluation time point.
#'
#' @return A single numeric value representing the Brier score a a specific time point.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   veteran <- survival::veteran
#'   mod <- survdnn(survival::Surv(time, status) ~
#'   age + karno + celltype, data = veteran, epochs = 50, verbose = FALSE)
#'   pred <- predict(mod, newdata = veteran, type = "survival", times = c(30, 90, 180))
#'   y <- model.response(model.frame(mod$formula, veteran))
#'   survdnn::brier(y, pre_sp = pred[["t=90"]], t_star = 90)
#' }
#' }

brier <- function(object, pre_sp, t_star) {
  if (!inherits(object, "Surv")) stop("object must be a survival object")
  if (length(pre_sp) != nrow(object)) stop("Length of predictions must match number of observations")

  time <- object[, 1]
  status <- object[, 2]
  km_fit <- survfit(Surv(time, 1 - status) ~ 1)

  all_times <- sort(unique(c(t_star, time)))
  G_all <- summary(km_fit, times = all_times, extend = TRUE)$surv
  names(G_all) <- as.character(all_times)

  Gt <- G_all[as.character(t_star)]

  if (is.na(Gt) || Gt == 0) {
    warning("brier: G(t_star) is NA or zero at t = ", t_star, "; returning NA.")
    return(NA_real_)
  }

  early_event <- which(time < t_star & status == 1)
  later_risk  <- which(time >= t_star)
  score_vec <- numeric(length(time))

  if (length(early_event) > 0) {
    Gti <- G_all[as.character(time[early_event])]
    valid <- which(!is.na(Gti) & Gti > 0)
    score_vec[early_event[valid]] <- (pre_sp[early_event[valid]]^2) / Gti[valid]
  }

  if (length(later_risk) > 0) {
    score_vec[later_risk] <- ((1 - pre_sp[later_risk])^2) / Gt
  }

  bs_value <- mean(score_vec, na.rm = TRUE)
  names(bs_value) <- "brier"
  return(round(bs_value, 6))
}


#' Integrated Brier Score (IBS) from a Survival Probability Matrix
#'
#' Computes the Integrated Brier Score (IBS) over a set of evaluation time points,
#' using trapezoidal integration and IPCW adjustment for right-censoring.
#'
#' @param object A `Surv` object with observed time and status.
#' @param sp_matrix A data frame or matrix of predicted survival probabilities.
#'   Each column corresponds to a time point in `times`.
#' @param times A numeric vector of time points. Must match the columns of `sp_matrix`.
#'
#' @return A single numeric value representing the integrated Brier score.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   set.seed(123)
#'   veteran <- survival::veteran
#'   idx <- sample(nrow(veteran), 0.7 * nrow(veteran))
#'   train <- veteran[idx, ]; test <- veteran[-idx, ]
#'   mod <- survdnn(survival::Surv(time, status) ~
#'   age + karno + celltype, data = train, epochs = 50, verbose = FALSE)
#'   pred <- predict(mod, newdata = test, times = c(30, 90, 180), type = "survival")
#'   y_test <- model.response(model.frame(mod$formula, test))
#'   ibs_survmat(y_test, sp_matrix = pred, times = c(30, 90, 180))
#' }
#' }

ibs_survmat <- function(object, sp_matrix, times) {
  if (!inherits(object, "Surv")) stop("object must be a survival object")
  if (length(times) != ncol(sp_matrix)) stop("Length of times must match sp_matrix columns")

  if (length(times) == 1) {
    return(brier(object, pre_sp = sp_matrix[, 1], t_star = times[1]))
  }

  times <- sort(times)
  brier_vec <- vapply(seq_along(times), function(i) {
    brier(object, pre_sp = sp_matrix[, i], t_star = times[i])
  }, numeric(1))

  dt <- diff(times)
  integral <- sum((brier_vec[-1] + brier_vec[-length(brier_vec)]) / 2 * dt)
  ibs_value <- integral / (max(times) - min(times))
  names(ibs_value) <- "ibs"
  return(round(ibs_value, 6))
}

test_that("cindex_survmat computes a valid C-index", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  times <- c(30, 90, 180)
  sp <- predict(mod, newdata = data, times = times, type = "survival")
  y <- model.response(model.frame(mod$formula, data))

  cidx <- cindex_survmat(y, sp, t_star = 90)
  expect_type(cidx, "double")
  expect_true(cidx >= 0 && cidx <= 1, info = paste("C-index out of bounds:", cidx))
})


test_that("brier returns a valid score", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  times <- 90
  sp <- predict(mod, newdata = data, times = times, type = "survival")
  y <- model.response(model.frame(mod$formula, data))

  bs <- brier(y, pre_sp = sp[["t=90"]], t_star = 90)
  expect_type(bs, "double")
  expect_true(bs >= 0 && bs <= 1, info = paste("Brier score out of bounds:", bs))
})


test_that("ibs_survmat returns a valid integrated Brier score", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  mod <- survdnn(Surv(time, status) ~ age + karno + celltype,
                 data = data, epochs = 5, verbose = FALSE)

  times <- c(30, 90, 180)
  sp <- predict(mod, newdata = data, times = times, type = "survival")
  y <- model.response(model.frame(mod$formula, data))

  ibs <- ibs_survmat(y, sp, times)
  expect_type(ibs, "double")
  expect_true(ibs >= 0 && ibs <= 1, info = paste("IBS out of bounds:", ibs))
})


test_that("metric functions handle dimension mismatches gracefully", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  y <- Surv(data$time, data$status)
  wrong_sp <- matrix(runif(200), nrow = 100)  # should be 137 rows

  expect_error(
    ibs_survmat(y, wrong_sp, times = c(30, 60)),
    regexp = "Length of predictions"
  )

  expect_error(
    brier(y, pre_sp = runif(100), t_star = 60),
    regexp = "Length of predictions"
  )
})

utils::globalVariables(c("epoch"))

#' Plot Training Loss for a survdnn Model
#'
#' Visualize the evolution of the training loss across epochs for a fitted
#' `survdnn` model. Helps inspect convergence, instability, or callback effects
#' (e.g., early stopping).
#'
#' @param object A fitted `survdnn` model.
#' @param smooth Logical; if `TRUE`, overlays a smoothed loess curve.
#' @param log_y Logical; if `TRUE`, uses a log10 y-scale.
#' @param ... Reserved for future use.
#'
#' @return A `ggplot` object.
#' @export
plot_loss <- function(object,
                      smooth = FALSE,
                      log_y  = FALSE,
                      ...) {
  stopifnot(inherits(object, "survdnn"))

  loss_history <- object$loss_history
  if (is.null(loss_history) || length(loss_history) == 0) {
    stop("Object has no `loss_history` to plot.", call. = FALSE)
  }

  df <- data.frame(
    epoch = seq_along(loss_history),
    loss  = as.numeric(loss_history)
  )

  p <- ggplot2::ggplot(df, ggplot2::aes(x = epoch, y = loss)) +
    ggplot2::geom_line() +
    ggplot2::labs(
      title = "survdnn training loss",
      x     = "Epoch",
      y     = "Loss"
    ) +
    ggplot2::theme_minimal()

  if (isTRUE(smooth))
    p <- p + ggplot2::geom_smooth(se = FALSE, method = "loess")

  if (isTRUE(log_y))
    p <- p + ggplot2::scale_y_log10()

  p
}

#' Early stopping callback for survdnn
#'
#' Simple early stopping on a monitored scalar, typically the training loss.
#'
#' @param patience Integer. Number of epochs with no improvement before stopping.
#' @param min_delta Minimum change to qualify as an improvement (default: 0).
#' @param mode Character. "min" (for losses) or "max" (for metrics to maximize).
#' @param verbose Logical. Whether to print a message when early stopping is triggered.
#'
#' @return A function of the form `function(epoch, current)` that returns TRUE
#'   if training should stop, FALSE otherwise.
#' @export
callback_early_stopping <- function(patience = 10L,
                                    min_delta = 0,
                                    mode = c("min", "max"),
                                    verbose = FALSE) {
  mode <- match.arg(mode)

  if (!is.numeric(patience) || patience < 1) {
    stop("`patience` must be a positive integer.", call. = FALSE)
  }

  best <- if (mode == "min") Inf else -Inf
  wait <- 0L

  function(epoch, current) {
    if (!is.numeric(current) || length(current) != 1L || is.na(current)) {
      warning("callback_early_stopping: `current` must be a single numeric value; ignoring callback for this epoch.")
      return(FALSE)
    }

    improved <- if (mode == "min") {
      current < (best - min_delta)
    } else {
      current > (best + min_delta)
    }

    if (improved) {
      best <<- current
      wait <<- 0L
      return(FALSE)
    }

    wait <<- wait + 1L
    if (wait >= patience) {
      if (verbose) {
        message(
          "Early stopping at epoch ", epoch,
          " (best = ", signif(best, 6), ", patience = ", patience, ")"
        )
      }
      return(TRUE)
    }

    FALSE
  }
}

#' Summarize a Deep Survival Neural Network Model
#'
#' Provides a structured summary of a fitted `survdnn` model, including the network architecture,
#' training configuration, and data characteristics. The summary is printed automatically with
#' a styled header and sectioned output using \{cli\} and base formatting. The object is returned invisibly.
#'
#' @param object An object of class `"survdnn"` returned by the [survdnn()] function.
#' @param ... Currently ignored (for future compatibility).
#'
#' @return Invisibly returns an object of class `"summary.survdnn"`.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   set.seed(42)
#'   sim_data <- data.frame(
#'     age = rnorm(100, 60, 10),
#'     sex = factor(sample(c("male", "female"), 100, TRUE)),
#'     trt = factor(sample(c("A", "B"), 100, TRUE)),
#'     time = rexp(100, 0.05),
#'     status = rbinom(100, 1, 0.7)
#'   )
#'   mod <- survdnn(
#'     survival::Surv(time, status) ~ age + sex + trt,
#'     data = sim_data,
#'     epochs = 50,
#'     verbose = FALSE
#'   )
#'   summary(mod)
#' }
#' }

summary.survdnn <- function(object, ...) {
  stopifnot(inherits(object, "survdnn"))

  # Rebuild model frame using the same NA policy as during fitting
  na_action <- if (!is.null(object$na_action)) object$na_action else "omit"
  mf <- model.frame(
    object$formula,
    data = object$data,
    na.action = if (na_action == "omit") stats::na.omit else stats::na.fail
  )

  y <- model.response(mf)
  time <- y[, "time"]
  status <- y[, "status"]

  n_before <- nrow(object$data)
  n_used   <- nrow(mf)
  n_removed <- n_before - n_used

  events   <- sum(status == 1, na.rm = TRUE)
  censored <- sum(status == 0, na.rm = TRUE)

  out <- list(
    model_summary = list(
      hidden_layers = object$hidden,
      activation    = object$activation,
      dropout       = if (!is.null(object$dropout)) object$dropout else NA_real_,
      batch_norm    = if (!is.null(object$batch_norm)) object$batch_norm else NA,
      final_loss    = object$final_loss,
      loss_history_length = if (!is.null(object$loss_history)) length(object$loss_history) else NA_integer_
    ),
    training_summary = list(
      epochs        = object$epochs,
      learning_rate = object$lr,
      loss_function = object$loss,
      optimizer     = object$optimizer,
      device        = object$device,
      threads       = if (!is.null(object$threads)) object$threads else NA_integer_,
      na_action     = object$na_action
    ),
    data_summary = list(
      observations_total = n_before,
      observations_used  = n_used,
      observations_removed = n_removed,
      predictors   = object$xnames,
      n_predictors = length(object$xnames),
      time_range   = range(time, na.rm = TRUE),
      events       = events,
      censored     = censored,
      event_rate   = mean(status == 1, na.rm = TRUE),
      standardized = !is.null(object$x_center) && !is.null(object$x_scale)
    ),
    loss_details = NULL,
    formula = object$formula
  )

  if (identical(object$loss, "aft")) {
    out$loss_details <- list(
      aft_loc       = object$aft_loc,
      aft_log_sigma = object$aft_log_sigma
    )
  } else if (identical(object$loss, "coxtime")) {
    out$loss_details <- list(
      coxtime_time_center = object$coxtime_time_center,
      coxtime_time_scale  = object$coxtime_time_scale
    )
  }

  class(out) <- "summary.survdnn"

  cli::cli_h1("Summary of survdnn model")

  cat("\nFormula:\n  ")
  print(out$formula)

  cat("\nModel architecture:\n")
  cat("  Hidden layers: ", paste(out$model_summary$hidden_layers, collapse = " : "), "\n")
  cat("  Activation: ", out$model_summary$activation, "\n")
  cat("  Dropout: ", out$model_summary$dropout, "\n")
  cat("  Batch norm: ", out$model_summary$batch_norm, "\n")
  cat("  Final loss: ", formatC(out$model_summary$final_loss, digits = 6, format = "f"), "\n")

  cat("\nTraining summary:\n")
  cat("  Epochs: ", out$training_summary$epochs, "\n")
  cat("  Learning rate: ", out$training_summary$learning_rate, "\n")
  cat("  Loss function: ", out$training_summary$loss_function, "\n")
  cat("  Optimizer: ", out$training_summary$optimizer, "\n")
  cat("  Device: ", as.character(out$training_summary$device), "\n")
  cat("  CPU threads: ", if (is.na(out$training_summary$threads)) "default" else out$training_summary$threads, "\n")
  cat("  NA action: ", as.character(out$training_summary$na_action), "\n")

  cat("\nData summary:\n")
  cat("  Observations (used/total): ", out$data_summary$observations_used, " / ",
      out$data_summary$observations_total, "\n", sep = "")
  if (out$data_summary$observations_removed > 0) {
    cat("  Removed due to missing values: ", out$data_summary$observations_removed, "\n", sep = "")
  }
  cat("  Predictors (", out$data_summary$n_predictors, "): ",
      paste(out$data_summary$predictors, collapse = ", "), "\n", sep = "")
  cat("  Time range: [", paste(out$data_summary$time_range, collapse = ", "), "]\n")
  cat("  Events / censored: ", out$data_summary$events, " / ", out$data_summary$censored, "\n")
  cat("  Event rate: ", sprintf("%.1f%%", 100 * out$data_summary$event_rate), "\n")
  cat("  Predictors standardized: ", if (isTRUE(out$data_summary$standardized)) "yes" else "no", "\n")

  if (!is.null(out$loss_details)) {
    cat("\nLoss-specific details:\n")
    for (nm in names(out$loss_details)) {
      cat("  ", nm, ": ", out$loss_details[[nm]], "\n", sep = "")
    }
  }

  invisible(out)
}

#' @import torch
#' @importFrom survival Surv survfit
#' @importFrom ggplot2 ggplot aes geom_line labs theme_minimal
#' @importFrom stats model.frame model.matrix model.response delete.response terms qnorm sd approx pnorm predict
#' @importFrom utils tail
#' @importFrom methods is
#' @importFrom dplyr all_of
NULL

utils::globalVariables(c("loss", "epoch"))

#' Tune Hyperparameters for a survdnn Model via Cross-Validation
#'
#' Performs k-fold cross-validation over a user-defined hyperparameter grid
#' and selects the best configuration according to the specified evaluation metric.
#'
#' @param formula A survival formula, e.g., `Surv(time, status) ~ x1 + x2`.
#' @param data A data frame.
#' @param times A numeric vector of evaluation time points.
#' @param metrics A character vector of evaluation metrics: "cindex", "brier", or "ibs".
#'   Only the first metric is used for model selection.
#' @param param_grid A named list defining hyperparameter combinations to evaluate.
#'   Required names: `hidden`, `lr`, `activation`, `epochs`, `loss`.
#' @param folds Number of cross-validation folds (default: 3).
#' @param .seed Optional seed for reproducibility.
#' @param .device Character string indicating the computation device used when fitting models
#'   during cross-validation and refitting. One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"`
#'   uses CUDA if available, otherwise falls back to CPU.
#' @param na_action Character. How to handle missing values:
#'   `"omit"` drops incomplete rows; `"fail"` errors if any NA is present.
#' @param refit Logical. If TRUE, refits the best model on the full dataset.
#' @param return One of "all", "summary", or "best_model":
#'   \describe{
#'     \item{"all"}{Returns the full cross-validation result across all combinations.}
#'     \item{"summary"}{Returns averaged results per configuration.}
#'     \item{"best_model"}{Returns the refitted model or best hyperparameters.}
#'   }
#'
#' @return A tibble or model object depending on the `return` value.
#' @export
tune_survdnn <- function(
  formula,
  data,
  times,
  metrics = "cindex",
  param_grid,
  folds = 3,
  .seed = 42,
  .device = c("auto", "cpu", "cuda"),
  na_action = c("omit", "fail"),
  refit = FALSE,
  return = c("all", "summary", "best_model")
) {
  return    <- match.arg(return)
  .device   <- match.arg(.device)
  na_action <- match.arg(na_action)

  if (!is.null(.seed)) survdnn_set_seed(.seed)

  param_df <- tidyr::crossing(!!!param_grid)

  all_results <- purrr::pmap_dfr(
    param_df,
    function(hidden, lr, activation, epochs, loss) {
      config_tbl <- tibble::tibble(
        hidden     = list(hidden),
        lr         = lr,
        activation = activation,
        epochs     = epochs,
        loss       = loss
      )

      cv_tbl <- cv_survdnn(
        formula    = formula,
        data       = data,
        times      = times,
        metrics    = metrics,
        folds      = folds,
        hidden     = hidden,
        lr         = lr,
        activation = activation,
        epochs     = epochs,
        loss       = loss,
        .seed      = .seed,
        .device    = .device,
        na_action  = na_action
      )

      dplyr::bind_cols(config_tbl[rep(1, nrow(cv_tbl)), ], cv_tbl)
    }
  )

  summary_tbl <- summarize_tune_survdnn(all_results, by_time = FALSE)

  ## select best hyperparameters
  primary_metric <- metrics[1]

  best_row_all <- all_results |>
    dplyr::filter(metric == primary_metric) |>
    dplyr::group_by(hidden, lr, activation, epochs, loss) |>
    dplyr::summarise(mean = mean(value, na.rm = TRUE), .groups = "drop") |>
    dplyr::slice_max(
      order_by = if (identical(primary_metric, "cindex")) mean else -mean,
      n = 1
    )

  if (nrow(best_row_all) == 0) {
    stop("No valid configuration found for primary metric: ", primary_metric, call. = FALSE)
  }

  ## refitting the best model
  if (refit) {
    message("Refitting best model on full data...")
    best_model <- survdnn(
      formula    = formula,
      data       = data,
      hidden     = best_row_all$hidden[[1]],
      lr         = best_row_all$lr,
      activation = best_row_all$activation,
      epochs     = best_row_all$epochs,
      loss       = best_row_all$loss[[1]],
      .seed      = .seed,
      .device    = .device,
      na_action  = na_action
    )
  }

  switch(
    return,
    "all"        = all_results,
    "summary"    = summary_tbl,
    "best_model" = if (refit) best_model else dplyr::select(best_row_all, -mean)
  )
}


#' Summarize survdnn Tuning Results
#'
#' Aggregates cross-validation results from `tune_survdnn(return = "all")`
#' by configuration, metric, and optionally by time point.
#'
#' @param tuning_results The full tibble returned by `tune_survdnn(..., return = "all")`.
#' @param by_time Logical; whether to group and summarize separately by time points.
#'
#' @return A summarized tibble with mean and standard deviation of performance metrics.
#' @export
summarize_tune_survdnn <- function(tuning_results, by_time = TRUE) {
  if (!all(c("metric", "value") %in% names(tuning_results))) {
    stop("Input must be the result of `tune_survdnn(return = 'all')`.", call. = FALSE)
  }

  group_vars <- c("hidden", "lr", "activation", "epochs", "loss", "metric")
  if (by_time && "time" %in% names(tuning_results)) {
    group_vars <- c(group_vars, "time")
  }

  tuning_results |>
    dplyr::group_by(dplyr::across(all_of(group_vars))) |>
    dplyr::summarise(
      mean = mean(value, na.rm = TRUE),
      sd   = sd(value, na.rm = TRUE),
      .groups = "drop"
    ) |>
    dplyr::arrange(metric, dplyr::desc(mean))
}

test_that("tune_survdnn returns correct structure for all modes", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  param_grid <- list(
    hidden     = list(c(8), c(8, 4)),
    lr         = c(1e-3),
    activation = c("relu"),
    epochs     = c(5),
    loss       = c("cox")
  )
  times <- c(30, 90)

  all_res <- tune_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = times,
    metrics = "cindex",
    param_grid = param_grid,
    folds = 2,
    .seed = 123,
    refit = FALSE,
    return = "all"
  )

  expect_s3_class(all_res, "data.frame")
  expect_true(all(c("fold", "metric", "value") %in% names(all_res)))

  summary_res <- tune_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = times,
    metrics = "cindex",
    param_grid = param_grid,
    folds = 2,
    .seed = 123,
    refit = FALSE,
    return = "summary"
  )

  expect_s3_class(summary_res, "data.frame")
  expect_true(all(c("metric", "mean", "sd") %in% names(summary_res)))

  best_cfg <- tune_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = times,
    metrics = "cindex",
    param_grid = param_grid,
    folds = 2,
    .seed = 123,
    refit = FALSE,
    return = "best_model"
  )

  expect_s3_class(best_cfg, "data.frame")
  expect_true(all(c("hidden", "lr", "activation", "epochs", "loss") %in% names(best_cfg)))
})

test_that("tune_survdnn works with refit = TRUE and returns survdnn model", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  param_grid <- list(
    hidden     = list(c(8)),
    lr         = c(1e-3),
    activation = c("relu"),
    epochs     = c(3),
    loss       = c("cox")
  )
  times <- c(30, 90)

  mod <- tune_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = times,
    metrics = "cindex",
    param_grid = param_grid,
    folds = 2,
    .seed = 42,
    refit = TRUE,
    return = "best_model"
  )

  expect_s3_class(mod, "survdnn")
  expect_s3_class(mod$model, "nn_module")
})

test_that("summarize_tune_survdnn aggregates correctly and throws on bad input", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran
  param_grid <- list(
    hidden     = list(c(8)),
    lr         = c(1e-3),
    activation = c("relu"),
    epochs     = c(3),
    loss       = c("cox")
  )
  times <- c(30, 90)

  all_res <- tune_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = times,
    metrics = "brier",  
    param_grid = param_grid,
    folds = 2,
    .seed = 123,
    refit = FALSE,
    return = "all"
  )

  sm <- summarize_tune_survdnn(all_res, by_time = TRUE)
  expect_s3_class(sm, "data.frame")

  if ("time" %in% names(all_res)) {
    expect_true("time" %in% names(sm))
  }

  expect_error(summarize_tune_survdnn(data.frame(a = 1)), "Input must be the result")
})

#' Evaluate a survdnn Model Using Survival Metrics
#'
#' Computes evaluation metrics for a fitted `survdnn` model at one or more time points.
#' Supported metrics include the concordance index (`"cindex"`), Brier score (`"brier"`),
#' and integrated Brier score (`"ibs"`).
#'
#' @param model A fitted `survdnn` model object.
#' @param metrics A character vector of metric names: `"cindex"`, `"brier"`, `"ibs"`.
#' @param times A numeric vector of evaluation time points.
#' @param newdata Optional. A data frame on which to evaluate the model. Defaults to training data.
#' @param na_action Character. How to handle missing values in evaluation data:
#'   `"omit"` drops incomplete rows, `"fail"` errors if any NA is present.
#' @param verbose Logical. If TRUE and `na_action="omit"`, prints a message when rows are removed.
#'
#' @return A tibble with evaluation results, containing at least `metric`, `value`, and possibly `time`.
#' @export
evaluate_survdnn <- function(model,
                             metrics = c("cindex", "brier", "ibs"),
                             times,
                             newdata = NULL,
                             na_action = c("omit", "fail"),
                             verbose = FALSE) {
  
  stopifnot(inherits(model, "survdnn"))
  if (missing(times)) stop("You must provide `times` for evaluation.", call. = FALSE)

  na_action <- match.arg(na_action)

  allowed_metrics <- c("cindex", "brier", "ibs")
  unknown <- setdiff(metrics, allowed_metrics)
  if (length(unknown) > 0) {
    stop("Unknown metric(s): ", paste(unknown, collapse = ", "), call. = FALSE)
  }
 
  data <- if (is.null(newdata)) model$data else newdata
  n_before <- nrow(data)
  
  # build model frame first with explicit NA policy
  mf <- model.frame(
    model$formula,
    data = data,
    na.action = if (na_action == "omit") stats::na.omit else stats::na.fail
  )
  
  n_after <- nrow(mf)
  n_removed <- n_before - n_after
  if (n_removed > 0 && isTRUE(verbose) && na_action == "omit") {
    message(sprintf("[survdnn::eval] removed %d observations with missing values in evaluation data.", n_removed))
  }

  y <- model.response(mf)
  if (!inherits(y, "Surv")) stop("The response must be a 'Surv' object.", call. = FALSE)

  # predict on the filtered mf to keep row alignment
  sp_matrix <- predict(model, newdata = mf, times = times, type = "survival")

  purrr::map_dfr(metrics, function(metric) { ## (to replace map_dfr() with fmap() from functionals package)
    if (metric == "brier" && length(times) > 1) {
      tibble::tibble(
        metric = "brier",
        time = times,
        value = vapply(seq_along(times), function(i) {
          brier(y, pre_sp = sp_matrix[, i], t_star = times[i])
        }, numeric(1))
      )
    } else {
      val <- switch(
        metric,
        "cindex" = cindex_survmat(y, predicted = sp_matrix, t_star = max(times)),
        "brier"  = brier(y, pre_sp = sp_matrix[, 1], t_star = times[1]),
        "ibs"    = ibs_survmat(y, sp_matrix, times)
      )
      tibble::tibble(metric = metric, value = val)
    }
  })
}


#' K-Fold Cross-Validation for survdnn Models
#'
#' Performs cross-validation for a `survdnn` model using the specified evaluation metrics.
#'
#' @param formula A survival formula, e.g., `Surv(time, status) ~ x1 + x2`.
#' @param data A data frame.
#' @param times A numeric vector of evaluation time points.
#' @param metrics A character vector: any of `"cindex"`, `"brier"`, `"ibs"`.
#' @param folds Integer. Number of folds to use.
#' @param .seed Optional. Set random seed for reproducibility.
#' @param .device Character string indicating the computation device used when fitting the models
#'   in each fold. One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"` uses CUDA if available,
#'   otherwise falls back to CPU.
#' @param .threads Optional positive integer. If provided, sets Torch CPU thread
#'   count before each fold fit via `torch::torch_set_num_threads()`.
#' @param na_action Character. How to handle missing values within each fold:
#'   `"omit"` drops incomplete rows; `"fail"` errors if any NA is present.
#' @param verbose Logical; whether to print cross-validation progress and propagate
#'   verbose messages to fitting/evaluation in each fold (default: TRUE).
#' @param ... Additional arguments passed to [survdnn()].
#'
#' @return A tibble containing metric values per fold and (optionally) per time point.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   veteran <- survival::veteran
#'   cv_survdnn(
#'     survival::Surv(time, status) ~ age + karno + celltype,
#'     data = veteran,
#'     times = c(30, 90, 180),
#'     metrics = "ibs",
#'     folds = 3,
#'     .seed = 42,
#'     hidden = c(16, 8),
#'     epochs = 5
#'   )
#' }
#' }
cv_survdnn <- function(formula, data, times,
                       metrics = c("cindex", "ibs"),
                       folds = 5,
                       .seed = NULL,
                       .device = c("auto", "cpu", "cuda"),
                       .threads = NULL,
                       na_action = c("omit", "fail"),
                       verbose = TRUE,
                       ...) {

  .device   <- match.arg(.device)
  na_action <- match.arg(na_action)

  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Package 'rsample' is required for cross-validation.", call. = FALSE)
  }

  if (!inherits(formula, "formula")) stop("`formula` must be a survival formula", call. = FALSE)
  if (!is.data.frame(data)) stop("`data` must be a data frame", call. = FALSE)
  if (missing(times)) stop("You must provide a `times` vector.", call. = FALSE)

  if (!is.null(.seed)) survdnn_set_seed(.seed)

  if (isTRUE(verbose)) {
    message(
      sprintf(
        "[survdnn::cv] start: folds=%d n=%d metrics=%s times=%s",
        folds,
        nrow(data),
        paste(metrics, collapse = ","),
        paste(times, collapse = ",")
      )
    )
    if (!is.null(.threads)) {
      message(sprintf("[survdnn::cv] cpu_threads=%d", as.integer(.threads)))
    }
  }

  status_var <- all.vars(formula[[2]])[2]          # more safe for extracting the status
  vfolds <- rsample::vfold_cv(data, v = folds, strata = dplyr::all_of(status_var))

  results <- purrr::imap_dfr(vfolds$splits, function(split, i) {

    ## re-seed inside every fold to ensure full reproducibility
    survdnn_set_seed(.seed)

    train_data <- rsample::analysis(split)
    test_data  <- rsample::assessment(split)

    if (isTRUE(verbose)) {
      message(
        sprintf(
          "[survdnn::cv] fold %d/%d: train_n=%d test_n=%d",
          i, folds, nrow(train_data), nrow(test_data)
        )
      )
    }

    model <- survdnn(
      formula,
      data      = train_data,
      verbose   = verbose,
      .seed     = .seed,
      .device   = .device,
      .threads  = .threads,
      na_action = na_action,
      ...
    )

    eval_tbl <- evaluate_survdnn(
      model,
      metrics   = metrics,
      times     = times,
      newdata   = test_data,
      na_action = na_action,
      verbose   = verbose
    )

    eval_tbl$fold <- i
    if (isTRUE(verbose)) {
      message(sprintf("[survdnn::cv] fold %d/%d done.", i, folds))
    }
    eval_tbl
  })

  if (isTRUE(verbose)) {
    message(sprintf("[survdnn::cv] done: completed %d folds.", folds))
  }

  dplyr::select(results, fold, metric, time = dplyr::any_of("time"), value)
}


#' Summarize Cross-Validation Results from survdnn
#'
#' Computes mean, standard deviation, and confidence intervals for metrics from cross-validation.
#'
#' @param cv_results A tibble returned by [cv_survdnn()].
#' @param by_time Logical. Whether to stratify results by `time` (if present).
#' @param conf_level Confidence level for the intervals (default: 0.95).
#'
#' @return A tibble summarizing mean, sd, and confidence bounds per metric (and per time if applicable).
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
#'   veteran <- survival::veteran
#'   res <- cv_survdnn(
#'     survival::Surv(time, status) ~ age + karno + celltype,
#'     data = veteran,
#'     times = c(30, 90, 180, 270),
#'     metrics = c("cindex", "ibs"),
#'     folds = 3,
#'     .seed = 42,
#'     hidden = c(16, 8),
#'     epochs = 5
#'   )
#'   summarize_cv_survdnn(res)
#' }
#' }
summarize_cv_survdnn <- function(cv_results, by_time = TRUE, conf_level = 0.95) {
  stopifnot(all(c("fold", "metric", "value") %in% names(cv_results)))

  z <- qnorm((1 + conf_level) / 2)
  group_vars <- if ("time" %in% names(cv_results) && by_time) {
    c("metric", "time")
  } else {
    "metric"
  }

  cv_results |>
    dplyr::group_by(dplyr::across(all_of(group_vars))) |>
    dplyr::summarize(
      mean = mean(value, na.rm = TRUE),
      sd = sd(value, na.rm = TRUE),
      n = dplyr::n(),
      se = sd / sqrt(n),
      lower = mean - z * se,
      upper = mean + z * se,
      .groups = "drop"
    ) |>
    dplyr::select(-n, -se)
}

test_that("cv_survdnn performs cross-validation and returns fold-level metrics", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran

  out <- cv_survdnn(
    Surv(time, status) ~ age + karno + celltype,
    data = data,
    times = c(30, 90),
    metrics = c("cindex", "ibs"),
    folds = 2,
    .seed = 123,
    .threads = 1,
    hidden = c(8),
    epochs = 5,
    verbose = FALSE
  )

  expect_s3_class(out, "data.frame")
  expect_true(all(c("fold", "metric", "value") %in% names(out)))
  expect_true(all(out$fold %in% 1:2))
})

test_that("cv_survdnn errors on missing inputs or bad arguments", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  df <- survival::veteran

  expect_error(
    cv_survdnn("not a formula", df, times = 50),
    "`formula` must be a survival formula"
  )

  expect_error(
    cv_survdnn(Surv(time, status) ~ ., "not a df", times = 50),
    "`data` must be a data frame"
  )

  expect_error(
    cv_survdnn(Surv(time, status) ~ ., df),
    "must provide a `times`"
  )
})

test_that("cv_survdnn verbose controls progress messages", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  data <- survival::veteran

  msg_on <- capture.output(
    cv_survdnn(
      Surv(time, status) ~ age + karno + celltype,
      data = data,
      times = c(30),
      metrics = c("ibs"),
      folds = 2,
      .seed = 123,
      .threads = 1,
      hidden = c(8),
      epochs = 1,
      verbose = TRUE
    ),
    type = "message"
  )

  expect_true(any(grepl("\\[survdnn::cv\\]", msg_on)))
  expect_true(any(grepl("\\[survdnn::fit\\]", msg_on)))

  msg_off <- capture.output(
    cv_survdnn(
      Surv(time, status) ~ age + karno + celltype,
      data = data,
      times = c(30),
      metrics = c("ibs"),
      folds = 2,
      .seed = 123,
      .threads = 1,
      hidden = c(8),
      epochs = 1,
      verbose = FALSE
    ),
    type = "message"
  )

  expect_false(any(grepl("\\[survdnn::cv\\]|\\[survdnn::fit\\]|\\[survdnn::eval\\]", msg_off)))
})

#' Loss Functions for survdnn Models
#'
#' These functions define various loss functions used internally by `survdnn()`
#' for training deep neural networks on right-censored survival data.
#'
#' @section Supported Losses:
#' - **Cox partial likelihood loss** (`cox_loss`): Negative partial log-likelihood used in proportional hazards modeling.
#' - **L2-penalized Cox loss** (`cox_l2_loss`): Adds L2 regularization to the Cox loss.
#' - **Accelerated Failure Time (AFT) loss** (`aft_loss`): Log-normal AFT **censored negative log-likelihood**
#'   (uses both events and censored observations).
#' - **CoxTime loss** (`coxtime_loss`): Placeholder (see details). A correct CoxTime loss requires access to the network and the full input tensor.
#'
#' @param pred A torch tensor of model predictions. Its interpretation depends
#'   on the loss function:
#'   \itemize{
#'     \item{\code{loss = "cox"} or \code{"cox_l2"}: linear predictors
#'           (log hazard ratios).}
#'     \item{\code{loss = "aft"}: predicted log survival times.}
#'     \item{\code{loss = "coxtime"}: predicted time-dependent risk scores.}
#'   }
#' @param true A tensor with two columns: observed time and status (1 = event, 0 = censored).
#' @param lambda Regularization parameter for `cox_l2_loss` (default: `1e-3`).
#' @param sigma Positive numeric scale parameter for the log-normal AFT model (default: `1`).
#'   In `survdnn()`, a learnable global scale can be used via `survdnn__aft_lognormal_nll_factory()`.
#' @param aft_loc Numeric scalar location offset for the AFT model on the log-time scale.
#'   When non-zero, the model is trained on centered log-times `log(time) - aft_loc` for better numerical stability.
#'   Prediction should add this offset back: `mu = mu_resid + aft_loc`.
#' @param eps Small constant for numerical stability (default: `1e-12`).
#'
#' @return A scalar `torch_tensor` representing the loss value.
#' @name survdnn_losses
#' @keywords internal
#' @examples
#' # Used internally by survdnn()
NULL


# internal utilities
#' @keywords internal
survdnn__zeros_like_scalar <- function(x) {
  torch::torch_zeros_like(x$view(c(1)))[1]
}

#' @keywords internal
survdnn__count_true <- function(mask) {
  as.integer(mask$sum()$item())
}

#' @keywords internal
survdnn__log_surv_std_normal <- function(z, eps = 1e-12) {
  sqrt2 <- torch::torch_sqrt(torch::torch_tensor(2, dtype = z$dtype, device = z$device))
  u     <- z / sqrt2
  S     <- torch::torch_clamp(0.5 * (1 - torch::torch_erf(u)), min = eps)
  torch::torch_log(S)
}


# Cox loss 

#' @rdname survdnn_losses
#' @export
cox_loss <- function(pred, true) {
  time   <- true[, 1]
  status <- true[, 2]

  idx <- torch::torch_argsort(time, descending = TRUE)
  status <- status[idx]

  lp <- -pred[idx, 1]

  event_mask <- (status == 1)
  if (survdnn__count_true(event_mask) == 0) {
    return(survdnn__zeros_like_scalar(lp[1]))
  }

  log_cumsum_exp <- torch::torch_logcumsumexp(lp, dim = 1)
  -torch::torch_mean(lp[event_mask] - log_cumsum_exp[event_mask])
}

#' @rdname survdnn_losses
#' @export
cox_l2_loss <- function(pred, true, lambda = 1e-3) {
  base_loss <- cox_loss(pred, true)
  lp <- -pred[, 1]
  l2_penalty <- lambda * torch::torch_mean(lp^2)
  base_loss + l2_penalty
}


# AFT loss (log-normal AFT censored negative log-likelihood)

#' @rdname survdnn_losses
#' @export
aft_loss <- function(pred, true, sigma = 1, aft_loc = 0, eps = 1e-12) {

  time   <- true[, 1]
  status <- true[, 2]

  t  <- torch::torch_clamp(time, min = eps)
  lt <- torch::torch_log(t)

  mu_resid <- pred[, 1]

  sigma_t <- torch::torch_tensor(
    as.numeric(sigma),
    dtype  = mu_resid$dtype,
    device = mu_resid$device
  )

  sigma_t <- torch::torch_clamp(sigma_t, min = eps)
  log_sigma <- torch::torch_log(sigma_t)

  aft_loc_t <- torch::torch_tensor(
    as.numeric(aft_loc),
    dtype  = mu_resid$dtype,
    device = mu_resid$device
  )
  
  lt_c <- lt - aft_loc_t

  # In the log-normal AFT model, log(T) = aft_loc + mu + sigma * Z,
  # with Z ~ N(0, 1). Here, `pred` represents the subject-specific
  # location term mu, while `sigma` controls global time dispersion.

  z <- (lt_c - mu_resid) / sigma_t
  
  logS <- survdnn__log_surv_std_normal(z, eps = eps)

  nll_event <- lt + log_sigma + 0.5 * z^2
  nll_cens  <- -logS

  nll <- torch::torch_where(status == 1, nll_event, nll_cens)
  torch::torch_mean(nll)
}



# CoxTime loss (not identifiable from pred alone)

#' @rdname survdnn_losses
#' @export
coxtime_loss <- function(pred, true) {
  stop(
    "coxtime_loss(pred, true) is not identifiable from `pred` alone.\n",
    "Cox-Time requires evaluating g(t_i, x_j) for all subjects j at each event time t_i.\n",
    "Use the internal factory `survdnn__coxtime_loss_factory()` from survdnn() where `net` and the full input tensor are available.",
    call. = FALSE
  )
}


# Internals (with  loss factory pattern)


# Correct CoxTime
# IMPORTANT FIX:
# - use `true[,1]` (RAW time) for sorting + risk sets
# - use `x_tensor[,1]` (TIME AS FED TO NET) when calling net

#' @keywords internal
survdnn__coxtime_loss_factory <- function(net) {

  force(net)

  function(x_tensor, true, eps = 1e-12) {

    time_raw <- true[, 1]
    status   <- true[, 2]
    n        <- time_raw$size()[[1]]

    d <- x_tensor$size()[[2]]
    if (d < 2) stop("CoxTime expects x_tensor with at least 2 columns: (time, x).", call. = FALSE)

    time_inp <- x_tensor[, 1]               # time as used by the net
    x_cov    <- x_tensor[, 2:d, drop = FALSE]

    ## sort by RAW time (risk sets depend on raw ordering. Not so precise but approximate the risk intensity)
    idx <- torch::torch_argsort(time_raw, descending = TRUE)

    time_raw <- time_raw[idx]
    time_inp <- time_inp[idx]
    status   <- status[idx]
    x_cov    <- x_cov[idx, , drop = FALSE]

    event_mask <- (status == 1)
    ne <- as.integer(event_mask$sum()$item())
    if (ne == 0) return(torch::torch_zeros_like(time_raw[1]))

    ## event times
    t_event_raw <- time_raw[event_mask]     # for risk sets
    t_event_inp <- time_inp[event_mask]     # for net input
    x_event     <- x_cov[event_mask, , drop = FALSE]

    ## numerator: g(t_i, x_i) for events
    inp_num <- torch::torch_cat(list(t_event_inp$unsqueeze(2), x_event), dim = 2)
    g_num   <- net(inp_num)[, 1]

    ## denominator: for each event time t_i, evaluate g(t_i, x_j) for all j
    p <- x_cov$size()[[2]]

    x_rep <- x_cov$unsqueeze(2)$expand(c(n, ne, p))$permute(c(2, 1, 3))
    t_rep <- t_event_inp$view(c(ne, 1, 1))$expand(c(ne, n, 1))  # time for net input
    inp_den <- torch::torch_cat(list(t_rep, x_rep), dim = 3)

    inp_den2 <- inp_den$reshape(c(ne * n, d))
    g_den2   <- net(inp_den2)[, 1]
    g_den    <- g_den2$reshape(c(ne, n))

    ## risk sets computed on RAW time
    time_j <- time_raw$view(c(1, n))
    t_i    <- t_event_raw$view(c(ne, 1))
    risk   <- (time_j >= t_i)

    neg_inf  <- torch::torch_tensor(-Inf, dtype = g_den$dtype, device = g_den$device)
    g_masked <- torch::torch_where(risk, g_den, neg_inf)

    log_denom <- torch::torch_logsumexp(g_masked, dim = 2)
    -torch::torch_mean(g_num - log_denom)
  }
}


# AFT log-normal censored NLL factory (learnable global log(sigma))
# with optional centering by aft_loc.

#' @keywords internal
survdnn__aft_lognormal_nll_factory <- function(device, aft_loc = 0) {

  log_sigma <- torch::torch_tensor(
    0,
    dtype = torch::torch_float(),
    device = device,
    requires_grad = TRUE
  )

  aft_loc_t <- torch::torch_tensor(
    as.numeric(aft_loc),
    dtype = torch::torch_float(),
    device = device
  )

  loss_fn <- function(net, x, y, eps = 1e-12) {

    time   <- y[, 1]
    status <- y[, 2]

    mu_resid <- net(x)[, 1]

    t  <- torch::torch_clamp(time, min = eps)
    lt <- torch::torch_log(t)

    lt_c <- lt - aft_loc_t

    sigma <- torch::torch_clamp(torch::torch_exp(log_sigma), min = eps)
    z <- (lt_c - mu_resid) / sigma

    logS <- survdnn__log_surv_std_normal(z, eps = eps)

    nll_event <- lt + log_sigma + 0.5 * z^2
    nll_cens  <- -logS

    nll <- torch::torch_where(status == 1, nll_event, nll_cens)
    torch::torch_mean(nll)
  }

  list(
    loss_fn = loss_fn,
    extra_params = list(log_sigma = log_sigma)
  )
}

test_that("cox_loss returns scalar tensor (and is differentiable)", {
  skip_if_not(torch_is_installed())

  time   <- torch_tensor(runif(20, 1, 100))
  status <- torch_tensor(sample(0:1, 20, replace = TRUE))
  y      <- torch_stack(list(time, status), dim = 2)

  # pred must require grad to test autograd
  pred <- torch_randn(20, 1, requires_grad = TRUE)

  loss <- cox_loss(pred, y)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(isTRUE(loss$requires_grad))

  expect_silent(loss$backward())
  expect_false(torch::is_undefined_tensor(pred$grad))
})

test_that("cox_l2_loss returns penalized scalar tensor (and is differentiable)", {
  skip_if_not(torch_is_installed())

  time   <- torch_tensor(runif(20, 1, 100))
  status <- torch_tensor(sample(0:1, 20, replace = TRUE))
  y      <- torch_stack(list(time, status), dim = 2)

  pred <- torch_randn(20, 1, requires_grad = TRUE)

  loss <- cox_l2_loss(pred, y, lambda = 0.01)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(isTRUE(loss$requires_grad))

  expect_silent(loss$backward())
  expect_false(torch::is_undefined_tensor(pred$grad))
})

test_that("aft_loss returns scalar tensor (censored log-normal NLL)", {
  skip_if_not(torch_is_installed())

  time   <- torch_tensor(runif(20, 1, 100))
  status <- torch_tensor(sample(0:1, 20, replace = TRUE))
  y      <- torch_stack(list(time, status), dim = 2)

  pred <- torch_randn(20, 1, requires_grad = TRUE)

  loss <- aft_loss(pred, y, sigma = 1)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(isTRUE(loss$requires_grad))

  expect_silent(loss$backward())
  expect_false(torch::is_undefined_tensor(pred$grad))
})

test_that("aft_loss is finite for fully censored data (no special-casing to 0)", {
  skip_if_not(torch_is_installed())

  time   <- torch_tensor(runif(10, 1, 100))
  status <- torch_zeros(10)
  y      <- torch_stack(list(time, status), dim = 2)

  pred <- torch_randn(10, 1, requires_grad = TRUE)

  loss <- aft_loss(pred, y, sigma = 1)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(is.finite(as.numeric(loss)))
  expect_gt(as.numeric(loss), 0)

  expect_silent(loss$backward())
  expect_false(torch::is_undefined_tensor(pred$grad))
})

test_that("aft_loss fully censored stays on same device as pred (CUDA)", {
  skip_if_not(torch_is_installed())
  skip_if_not(cuda_is_available())

  device <- torch_device("cuda")

  time   <- torch_tensor(runif(10, 1, 100), device = device)
  status <- torch_zeros(10, device = device)
  y      <- torch_stack(list(time, status), dim = 2)

  pred <- torch_randn(10, 1, device = device, requires_grad = TRUE)

  loss <- aft_loss(pred, y, sigma = 1)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(is.finite(as.numeric(loss$to(device = "cpu"))))
  expect_equal(loss$device$type, pred$device$type)

  expect_silent(loss$backward())
  expect_false(torch::is_undefined_tensor(pred$grad))
})

test_that("coxtime_loss() is a placeholder and errors (use factory in training)", {
  skip_if_not(torch_is_installed())

  n <- 50
  pred   <- torch_randn(c(n, 1))
  time   <- torch_rand(c(n, 1)) * 100
  status <- torch_ones(c(n, 1))
  true   <- torch_cat(list(time, status), dim = 2)

  expect_error(
    coxtime_loss(pred, true),
    regexp = "not identifiable|requires evaluating",
    fixed  = FALSE
  )
})

test_that("survdnn__coxtime_loss_factory returns scalar tensor and is differentiable", {
  skip_if_not(torch_is_installed())

  n <- 30
  p <- 4
  d <- 1 + p

  # simple net; no BN to keep this unit test lightweight
  net <- torch::nn_sequential(
    torch::nn_linear(d, 8),
    torch::nn_relu(),
    torch::nn_linear(8, 1)
  )
  net$train()

  time   <- torch_tensor(runif(n, 1, 100))
  status <- torch_tensor(sample(0:1, n, replace = TRUE))
  y      <- torch_stack(list(time, status), dim = 2)

  x_cov  <- torch_randn(n, p)
  x_tensor <- torch_cat(list(time$unsqueeze(2), x_cov), dim = 2) # [n, 1+p]

  lf <- survdnn__coxtime_loss_factory(net)
  loss <- lf(x_tensor, y)

  expect_true(inherits(loss, "torch_tensor"))
  expect_equal(loss$numel(), 1)
  expect_true(isTRUE(loss$requires_grad))

  expect_silent(loss$backward())

  # at least one parameter should have a defined gradient
  grads <- lapply(net$parameters, function(par) par$grad)
  gsum <- sum(vapply(grads, function(g) if (is.null(g)) 0 else as.numeric(g$abs()$sum()$item()), numeric(1)))
  expect_gt(gsum, 0)
})

#' Predict from a survdnn Model
#'
#' Generate predictions from a fitted `survdnn` model for new data. Supports linear predictors,
#' survival probabilities at specified time points, or cumulative risk estimates.
#'
#' @param object An object of class `"survdnn"` returned by [survdnn()].
#' @param newdata A data frame of new observations to predict on.
#' @param times Numeric vector of time points at which to compute survival or risk probabilities.
#'   Required if `type = "survival"` or `type = "risk"` for Cox/AFT models.
#'   For CoxTime, `times = NULL` is allowed when `type="survival"` and defaults to event times.
#' @param type Character string specifying the type of prediction to return:
#'   \describe{
#'     \item{"lp"}{Linear predictor. For `"cox"`/`"cox_l2"` this is a log-risk score
#'     (higher implies worse prognosis, consistent with training sign convention). For `"aft"`,
#'     this is the predicted location parameter \eqn{\mu(x)} on the log-time scale. For `"coxtime"`,
#'     this is \eqn{g(t_0, x)} evaluated at a reference time \eqn{t_0} (the first event time).}
#'     \item{"survival"}{Predicted survival probabilities at each value of `times`.}
#'     \item{"risk"}{Cumulative risk (1 - survival) at **a single** time point.}
#'   }
#' @param ... Currently ignored (for future extensions).
#'
#' @return A numeric vector (if `type = "lp"` or `"risk"`), or a data frame
#'   (if `type = "survival"`) with one row per observation and one column per `times`.
#'
#' @export
predict.survdnn <- function(
  object,
  newdata,
  times = NULL,
  type = c("survival", "lp", "risk"),
  ...
) {

  type <- match.arg(type)

  stopifnot(inherits(object, "survdnn"))
  stopifnot(is.data.frame(newdata))

  loss  <- object$loss
  model <- object$model

  device <- if (!is.null(object$device)) object$device else torch::torch_device("cpu")
  model$to(device = device)
  model$eval()

  ## expand '.' safely using training data
  tt <- stats::terms(object$formula, data = object$data)

  x <- stats::model.matrix(
    stats::delete.response(tt),
    newdata
  )[, object$xnames, drop = FALSE]

  x_scaled <- scale(
    x,
    center = object$x_center,
    scale  = object$x_scale
  )

  ## IMPORTANT: type='risk' is defined at a single time point
  if (type == "risk" && !is.null(times) && length(times) != 1) {
    stop("For type = 'risk', `times` must be a single numeric value.", call. = FALSE)
  }

  ## Cox / Cox L2
  if (loss %in% c("cox", "cox_l2")) {

    if (type %in% c("survival", "risk") && is.null(times)) {
      stop("`times` must be specified for type = 'survival' or 'risk'.", call. = FALSE)
    }

    x_tensor <- torch::torch_tensor(
      x_scaled,
      dtype  = torch::torch_float(),
      device = device
    )

    torch::with_no_grad({
      lp <- -as.numeric(model(x_tensor)[, 1])
    })

    if (type == "lp") return(lp)

    ## baseline hazard via Breslow on training data
    train_x <- stats::model.matrix(
      stats::delete.response(tt),
      object$data
    )[, object$xnames, drop = FALSE]

    train_x_scaled <- scale(
      train_x,
      center = object$x_center,
      scale  = object$x_scale
    )

    train_x_tensor <- torch::torch_tensor(
      train_x_scaled,
      dtype  = torch::torch_float(),
      device = device
    )

    torch::with_no_grad({
      train_lp <- -as.numeric(model(train_x_tensor)[, 1])
    })

    y_train <- model.response(model.frame(object$formula, object$data))

    train_df <- data.frame(
      time   = y_train[, "time"],
      status = y_train[, "status"],
      lp     = train_lp
    )

    bh <- survival::basehaz(
      survival::coxph(survival::Surv(time, status) ~ lp, data = train_df),
      centered = FALSE
    )

    times_sorted <- sort(as.numeric(times))
    H0 <- stats::approx(bh$time, bh$hazard, xout = times_sorted, rule = 2)$y

    surv_mat <- outer(
      lp,
      H0,
      function(lp_i, h0_j) exp(-h0_j * exp(lp_i))
    )

    if (type == "risk") return(1 - surv_mat[, 1])

    colnames(surv_mat) <- paste0("t=", times_sorted)
    return(as.data.frame(surv_mat))
  }

  ## AFT (log-normal AFT with learned global sigma + training centering)
  if (loss == "aft") {

    if (type %in% c("survival", "risk") && is.null(times)) {
      stop("`times` must be specified for type = 'survival' or 'risk'.", call. = FALSE)
    }

    x_tensor <- torch::torch_tensor(
      x_scaled,
      dtype  = torch::torch_float(),
      device = device
    )

    torch::with_no_grad({
      mu_raw <- as.numeric(model(x_tensor)[, 1])
    })

    loc <- if (!is.null(object$aft_loc) && is.finite(object$aft_loc)) object$aft_loc else 0
    mu  <- mu_raw + loc
    if (type == "lp") return(mu)

    ## sigma: must be finite and > 0; otherwise fall back to 1
    ls <- object$aft_log_sigma
    sigma <- if (!is.null(ls) && is.finite(ls)) exp(ls) else 1
    if (!is.finite(sigma) || sigma <= 0) sigma <- 1

    times_sorted <- sort(as.numeric(times))
    times_sorted[times_sorted <= 0] <- .Machine$double.eps
    logt <- log(times_sorted)

    surv_mat <- outer(
      mu,
      logt,
      function(mu_i, lt) 1 - stats::pnorm((lt - mu_i) / sigma)
    )

    surv_mat[surv_mat < 0] <- 0
    surv_mat[surv_mat > 1] <- 1

    if (type == "risk") return(1 - surv_mat[, 1])

    colnames(surv_mat) <- paste0("t=", times_sorted)
    return(as.data.frame(surv_mat))
  }

  ## CoxTime
  if (loss == "coxtime") {

    y_train <- model.response(model.frame(object$formula, object$data))
    time_train   <- y_train[, "time"]
    status_train <- y_train[, "status"]
    event_times  <- sort(unique(time_train[status_train == 1]))

    if (length(event_times) == 0) {
      stop("CoxTime prediction requires at least one event in training data.", call. = FALSE)
    }

    ## --- time scaling used in training (fallback-safe) ---
    t_center <- if (!is.null(object$coxtime_time_center) && is.finite(object$coxtime_time_center)) {
      as.numeric(object$coxtime_time_center)
    } else 0

    t_scale <- if (!is.null(object$coxtime_time_scale) && is.finite(object$coxtime_time_scale) &&
                  as.numeric(object$coxtime_time_scale) > 0) {
      as.numeric(object$coxtime_time_scale)
    } else 1

    scale_t <- function(t) (as.numeric(t) - t_center) / t_scale

    ## training covariates (scaled) for baseline computation
    train_x <- stats::model.matrix(
      stats::delete.response(tt),
      object$data
    )[, object$xnames, drop = FALSE]

    train_x_scaled <- scale(
      train_x,
      center = object$x_center,
      scale  = object$x_scale
    )

    ## type = "lp": define lp at a reference time (first event time)
    if (type == "lp") {
      t0 <- event_times[1]
      t0s <- scale_t(t0)
      x_temp <- cbind(t0s, x_scaled)
      x_tensor <- torch::torch_tensor(x_temp, dtype = torch::torch_float(), device = device)
      torch::with_no_grad({
        lp <- as.numeric(model(x_tensor)[, 1])
      })
      return(lp)
    }

    ## For CoxTime: allow times=NULL for survival -> default event_times (RAW)
    if (type == "survival" && is.null(times)) {
      times_sorted <- event_times
    } else {
      if (type %in% c("survival", "risk") && is.null(times)) {
        stop("`times` must be specified for type = 'survival' or 'risk'.", call. = FALSE)
      }
      times_sorted <- sort(unique(as.numeric(times)))
    }

    ## Compute g(t_k, x_new) on event-time grid
    ## NOTE: net expects SCALED time input
    g_new_mat <- matrix(NA_real_, nrow = nrow(x_scaled), ncol = length(event_times))
    for (j in seq_along(event_times)) {
      tj  <- event_times[j]
      tjs <- scale_t(tj)
      x_temp <- cbind(tjs, x_scaled)
      x_tensor <- torch::torch_tensor(x_temp, dtype = torch::torch_float(), device = device)
      torch::with_no_grad({
        g_new_mat[, j] <- as.numeric(model(x_tensor)[, 1])
      })
    }

    ## Compute g(t_k, x_train) on event-time grid (scaled time input)
    g_train_mat <- matrix(NA_real_, nrow = nrow(train_x_scaled), ncol = length(event_times))
    for (j in seq_along(event_times)) {
      tj  <- event_times[j]
      tjs <- scale_t(tj)
      x_temp <- cbind(tjs, train_x_scaled)
      x_tensor <- torch::torch_tensor(x_temp, dtype = torch::torch_float(), device = device)
      torch::with_no_grad({
        g_train_mat[, j] <- as.numeric(model(x_tensor)[, 1])
      })
    }

    ## Baseline increments: dH0(t_k) = dN(t_k) / sum_{j in R(t_k)} exp(g(t_k, x_j))
    ## risk sets are defined on RAW time (a proxy for risk intensity)
    dN <- as.numeric(table(factor(time_train[status_train == 1], levels = event_times)))

    risk_mat <- outer(time_train, event_times, `>=`)
    denom <- colSums(exp(g_train_mat) * risk_mat, na.rm = TRUE)

    denom[denom <= 0] <- NA_real_
    dH0 <- dN / denom
    dH0[is.na(dH0)] <- 0

    ## Cumulative hazard at requested times (RAW time grid)
    H_pred <- matrix(0, nrow = nrow(g_new_mat), ncol = length(times_sorted))
    for (i in seq_along(times_sorted)) {
      relevant <- which(event_times <= times_sorted[i])
      if (length(relevant) > 0) {
        H_pred[, i] <- rowSums(
          exp(g_new_mat[, relevant, drop = FALSE]) *
            matrix(rep(dH0[relevant], each = nrow(g_new_mat)), nrow = nrow(g_new_mat))
        )
      } else {
        H_pred[, i] <- 0
      }
    }

    S_pred <- exp(-H_pred)
    S_pred[S_pred < 0] <- 0
    S_pred[S_pred > 1] <- 1

    if (type == "risk") return(1 - S_pred[, 1])

    colnames(S_pred) <- paste0("t=", times_sorted)
    return(as.data.frame(S_pred))
  }

  stop("Unsupported loss type for prediction: ", loss, call. = FALSE)
}

#' Print a survdnn Model
#'
#' Pretty prints a fitted `survdnn` model. Displays the formula,
#' network architecture, training configuration, and final training loss.
#'
#' @param x An object of class `"survdnn"`, returned by [survdnn()].
#' @param ... Ignored (for future compatibility).
#'
#' @return The model object, invisibly.
#' @export
#'
#' @examples
#' \donttest{
#' library(survival)
#' data(veteran, package = "survival")
#' mod <- survdnn(Surv(time, status) ~
#' age + karno + celltype, data = veteran, epochs = 20, verbose = FALSE)
#' print(mod)
#' }

print.survdnn <- function(x, ...) {
  stopifnot(inherits(x, "survdnn"))

  cli::cli_h1("survdnn model")
  cli::cli_text("Formula: {.val {deparse(x$formula)}}")
  cli::cli_text("Hidden layers: {.val {paste(x$hidden, collapse = ' : ')}}")
  cli::cli_text("Activation: {.val {x$activation}}")
  cli::cli_text("Learning rate: {.val {x$lr}}")

  if (!is.null(x$optimizer)) {
    cli::cli_text("Optimizer: {.val {x$optimizer}}")
  }

  cli::cli_text("Loss function: {.val {x$loss}}")

  if (is.numeric(x$final_loss) && length(x$final_loss) == 1 && !is.na(x$final_loss)) {
    cli::cli_text("Final loss: {.val {round(x$final_loss, 4)}} after {.val {x$epochs}} epochs")
  } else {
    cli::cli_alert_warning("Final loss is not available.")
  }

  invisible(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{brier}
\alias{brier}
\title{Brier Score for Right-Censored Survival Data at a Fixed Time}
\usage{
brier(object, pre_sp, t_star)
}
\arguments{
\item{object}{A `Surv` object with observed time and status.}

\item{pre_sp}{A numeric vector of predicted survival probabilities at `t_star`.}

\item{t_star}{The evaluation time point.}
}
\value{
A single numeric value representing the Brier score a a specific time point.
}
\description{
Computes the Brier score at a fixed time point using inverse probability of censoring weights (IPCW).
}
\examples{
\donttest{
library(survival)
data(veteran, package = "survival")
mod <- survdnn(Surv(time, status) ~
age + karno + celltype, data = veteran, epochs = 50, verbose = FALSE)
pred <- predict(mod, newdata = veteran, type = "survival", times = c(30, 90, 180))
y <- model.response(model.frame(mod$formula, veteran))
survdnn::brier(y, pre_sp = pred[["t=90"]], t_star = 90)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survdnn.R
\name{survdnn}
\alias{survdnn}
\title{Fit a Deep Neural Network for Survival Analysis}
\usage{
survdnn(
  formula,
  data,
  hidden = c(32L, 16L),
  activation = "relu",
  lr = 1e-04,
  epochs = 300L,
  loss = c("cox", "cox_l2", "aft", "coxtime"),
  optimizer = c("adam", "adamw", "sgd", "rmsprop", "adagrad"),
  optim_args = list(),
  verbose = TRUE,
  dropout = 0.3,
  batch_norm = TRUE,
  callbacks = NULL,
  .seed = NULL,
  .device = c("auto", "cpu", "cuda"),
  na_action = c("omit", "fail")
)
}
\arguments{
\item{formula}{A survival formula of the form `Surv(time, status) ~ predictors`.}

\item{data}{A data frame containing the variables in the model.}

\item{hidden}{Integer vector. Sizes of the hidden layers (default: c(32, 16)).}

\item{activation}{Character string specifying the activation function to use in each layer.
Supported options: `"relu"`, `"leaky_relu"`, `"tanh"`, `"sigmoid"`, `"gelu"`, `"elu"`, `"softplus"`.}

\item{lr}{Learning rate for the optimizer (default: `1e-4`).}

\item{epochs}{Number of training epochs (default: 300).}

\item{loss}{Character name of the loss function to use. One of `"cox"`, `"cox_l2"`, `"aft"`, or `"coxtime"`.}

\item{optimizer}{Character string specifying the optimizer to use. One of
`"adam"`, `"adamw"`, `"sgd"`, `"rmsprop"`, or `"adagrad"`. Defaults to `"adam"`.}

\item{optim_args}{Optional named list of additional arguments passed to the
underlying torch optimizer (e.g., `list(weight_decay = 1e-4, momentum = 0.9)`).}

\item{verbose}{Logical; whether to print loss progress every 50 epochs (default: TRUE).}

\item{dropout}{Numeric between 0 and 1. Dropout rate applied after each
hidden layer (default = 0.3). Set to 0 to disable dropout.}

\item{batch_norm}{Logical; whether to add batch normalization after each
hidden linear layer (default = TRUE).}

\item{callbacks}{Optional list of callback functions. Each callback should have
signature `function(epoch, current)` and return TRUE if training should stop,
FALSE otherwise. Used, for example, with [callback_early_stopping()].}

\item{.seed}{Optional integer. If provided, sets both R and torch random seeds for reproducible
weight initialization, shuffling, and dropout.}

\item{.device}{Character string indicating the computation device.
One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"` uses CUDA if available,
otherwise falls back to CPU.}

\item{na_action}{Character. How to handle missing values in the model variables:
`"omit"` drops incomplete rows (and reports how many were removed when `verbose=TRUE`);
`"fail"` stops with an error if any missing values are present.}
}
\value{
An object of class `"survdnn"` containing:
\describe{
  \item{model}{Trained `nn_module` object.}
  \item{formula}{Original survival formula.}
  \item{data}{Training data used for fitting.}
  \item{xnames}{Predictor variable names.}
  \item{x_center}{Column means of predictors.}
  \item{x_scale}{Column standard deviations of predictors.}
  \item{loss_history}{Vector of loss values per epoch.}
  \item{final_loss}{Final training loss.}
  \item{loss}{Loss function name used ("cox", "aft", etc.).}
  \item{activation}{Activation function used.}
  \item{hidden}{Hidden layer sizes.}
  \item{lr}{Learning rate.}
  \item{epochs}{Number of training epochs.}
  \item{optimizer}{Optimizer name used.}
  \item{optim_args}{List of optimizer arguments used.}
  \item{device}{Torch device used for training (`torch_device`).}
  \item{aft_log_sigma}{Learned global log(sigma) for `loss="aft"`; `NA_real_` otherwise.}
  \item{aft_loc}{AFT log-time location offset used for centering when `loss="aft"`; `NA_real_` otherwise.}
  \item{coxtime_time_center}{Mean used to scale time for CoxTime; `NA_real_` otherwise.}
  \item{coxtime_time_scale}{SD used to scale time for CoxTime; `NA_real_` otherwise.}
}
}
\description{
Trains a deep neural network (DNN) to model right-censored survival data
using one of the predefined loss functions: Cox, AFT, or Coxtime.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluation.R
\name{summarize_cv_survdnn}
\alias{summarize_cv_survdnn}
\title{Summarize Cross-Validation Results from survdnn}
\usage{
summarize_cv_survdnn(cv_results, by_time = TRUE, conf_level = 0.95)
}
\arguments{
\item{cv_results}{A tibble returned by [cv_survdnn()].}

\item{by_time}{Logical. Whether to stratify results by `time` (if present).}

\item{conf_level}{Confidence level for the intervals (default: 0.95).}
}
\value{
A tibble summarizing mean, sd, and confidence bounds per metric (and per time if applicable).
}
\description{
Computes mean, standard deviation, and confidence intervals for metrics from cross-validation.
}
\examples{
\donttest{
if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
  veteran <- survival::veteran
  res <- cv_survdnn(
    survival::Surv(time, status) ~ age + karno + celltype,
    data = veteran,
    times = c(30, 90, 180, 270),
    metrics = c("cindex", "ibs"),
    folds = 3,
    .seed = 42,
    hidden = c(16, 8),
    epochs = 5
  )
  summarize_cv_survdnn(res)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_survdnn.R
\name{summarize_tune_survdnn}
\alias{summarize_tune_survdnn}
\title{Summarize survdnn Tuning Results}
\usage{
summarize_tune_survdnn(tuning_results, by_time = TRUE)
}
\arguments{
\item{tuning_results}{The full tibble returned by `tune_survdnn(..., return = "all")`.}

\item{by_time}{Logical; whether to group and summarize separately by time points.}
}
\value{
A summarized tibble with mean and standard deviation of performance metrics.
}
\description{
Aggregates cross-validation results from `tune_survdnn(return = "all")`
by configuration, metric, and optionally by time point.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridsearch_survdnn.R
\name{gridsearch_survdnn}
\alias{gridsearch_survdnn}
\title{Grid Search for survdnn Hyperparameters}
\usage{
gridsearch_survdnn(
  formula,
  train,
  valid,
  times,
  metrics = c("cindex", "ibs"),
  param_grid,
  .seed = 42,
  .device = c("auto", "cpu", "cuda")
)
}
\arguments{
\item{formula}{A survival formula (e.g., `Surv(time, status) ~ .`)}

\item{train}{Training dataset}

\item{valid}{Validation dataset}

\item{times}{Evaluation time points (numeric vector)}

\item{metrics}{Evaluation metrics (character vector): any of "cindex" and "ibs".}

\item{param_grid}{A named list of hyperparameters to search over. Currently
supported entries are \code{hidden}, \code{lr}, \code{activation},
\code{epochs}, and \code{loss}.}

\item{.seed}{Optional random seed for reproducibility}

\item{.device}{Character string indicating the computation device used when
fitting all models in the grid search. One of \code{"auto"}, \code{"cpu"},
or \code{"cuda"}. This is a runtime setting and is not part of the
hyperparameter grid.}
}
\value{
A tibble with configurations and their validation metrics
}
\description{
Performs grid search over user-specified hyperparameters and evaluates performance on a validation set.
}
\examples{
\donttest{
if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
  set.seed(123)

  # Simulate small dataset
  n <- 60
  x1 <- rnorm(n); x2 <- rbinom(n, 1, 0.5)
  time <- rexp(n, rate = 0.1)
  status <- rbinom(n, 1, 0.7)
  df <- data.frame(time, status, x1, x2)

  # Split into training and validation
  idx <- sample(seq_len(n), 0.7 * n)
  train <- df[idx, ]
  valid <- df[-idx, ]

  # Define formula and param grid
  formula <- survival::Surv(time, status) ~ x1 + x2
  param_grid <- list(
    hidden     = list(c(4)),
    lr         = c(1e-3),
    activation = c("relu"),
    epochs     = c(1),
    loss       = c("cox")
  )

  # Run grid search
  results <- gridsearch_survdnn(
    formula = formula,
    train   = train,
    valid   = valid,
    times   = c(10, 20),
    metrics = c("cindex", "ibs"),
    param_grid = param_grid
  )

  # View summary
  dplyr::group_by(results, hidden, lr, activation, epochs, loss, metric) |>
    dplyr::summarise(mean = mean(value, na.rm = TRUE), .groups = "drop")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/losses.R
\name{survdnn_losses}
\alias{survdnn_losses}
\alias{cox_loss}
\alias{cox_l2_loss}
\alias{aft_loss}
\alias{coxtime_loss}
\title{Loss Functions for survdnn Models}
\usage{
cox_loss(pred, true)

cox_l2_loss(pred, true, lambda = 0.001)

aft_loss(pred, true, sigma = 1, aft_loc = 0, eps = 1e-12)

coxtime_loss(pred, true)
}
\arguments{
\item{pred}{A torch tensor of model predictions. Its interpretation depends
on the loss function:
\itemize{
  \item{\code{loss = "cox"} or \code{"cox_l2"}: linear predictors
        (log hazard ratios).}
  \item{\code{loss = "aft"}: predicted log survival times.}
  \item{\code{loss = "coxtime"}: predicted time-dependent risk scores.}
}}

\item{true}{A tensor with two columns: observed time and status (1 = event, 0 = censored).}

\item{lambda}{Regularization parameter for `cox_l2_loss` (default: `1e-3`).}

\item{sigma}{Positive numeric scale parameter for the log-normal AFT model (default: `1`).
In `survdnn()`, a learnable global scale can be used via `survdnn__aft_lognormal_nll_factory()`.}

\item{aft_loc}{Numeric scalar location offset for the AFT model on the log-time scale.
When non-zero, the model is trained on centered log-times `log(time) - aft_loc` for better numerical stability.
Prediction should add this offset back: `mu = mu_resid + aft_loc`.}

\item{eps}{Small constant for numerical stability (default: `1e-12`).}
}
\value{
A scalar `torch_tensor` representing the loss value.
}
\description{
These functions define various loss functions used internally by `survdnn()`
for training deep neural networks on right-censored survival data.
}
\section{Supported Losses}{

- **Cox partial likelihood loss** (`cox_loss`): Negative partial log-likelihood used in proportional hazards modeling.
- **L2-penalized Cox loss** (`cox_l2_loss`): Adds L2 regularization to the Cox loss.
- **Accelerated Failure Time (AFT) loss** (`aft_loss`): Log-normal AFT **censored negative log-likelihood**
  (uses both events and censored observations).
- **CoxTime loss** (`coxtime_loss`): Placeholder (see details). A correct CoxTime loss requires access to the network and the full input tensor.
}

\examples{
# Used internally by survdnn()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.survdnn.R
\name{plot.survdnn}
\alias{plot.survdnn}
\title{Plot survdnn Survival Curves using ggplot2}
\usage{
\method{plot}{survdnn}(
  x,
  newdata = NULL,
  times = 1:365,
  group_by = NULL,
  plot_mean_only = FALSE,
  add_mean = TRUE,
  alpha = 0.3,
  mean_lwd = 1.3,
  mean_lty = 1,
  ...
)
}
\arguments{
\item{x}{A fitted `survdnn` model object.}

\item{newdata}{Optional data frame for prediction (defaults to training data).}

\item{times}{A numeric vector of time points at which to compute survival probabilities.}

\item{group_by}{Optional name of a column in `newdata` used to color and group curves.}

\item{plot_mean_only}{Logical; if `TRUE`, plots only the mean survival curve per group.}

\item{add_mean}{Logical; if `TRUE`, adds mean curves to the individual lines.}

\item{alpha}{Alpha transparency for individual curves (ignored if `plot_mean_only = TRUE`).}

\item{mean_lwd}{Line width for mean survival curves.}

\item{mean_lty}{Line type for mean survival curves.}

\item{...}{Reserved for future use.}
}
\value{
A `ggplot` object.
}
\description{
Visualizes survival curves predicted by a fitted `survdnn` model.
Curves can be grouped by a categorical variable in `newdata` and
optionally display only the group-wise means or overlay them.
}
\examples{
\donttest{
if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
  set.seed(42)
  veteran <- survival::veteran
  mod <- survdnn(survival::Surv(time, status) ~ age + karno + celltype, data = veteran,
                 hidden = c(16, 8), epochs = 100, verbose = FALSE)
  plot(mod, group_by = "celltype", times = 1:300)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluation.R
\name{evaluate_survdnn}
\alias{evaluate_survdnn}
\title{Evaluate a survdnn Model Using Survival Metrics}
\usage{
evaluate_survdnn(
  model,
  metrics = c("cindex", "brier", "ibs"),
  times,
  newdata = NULL,
  na_action = c("omit", "fail"),
  verbose = FALSE
)
}
\arguments{
\item{model}{A fitted `survdnn` model object.}

\item{metrics}{A character vector of metric names: `"cindex"`, `"brier"`, `"ibs"`.}

\item{times}{A numeric vector of evaluation time points.}

\item{newdata}{Optional. A data frame on which to evaluate the model. Defaults to training data.}

\item{na_action}{Character. How to handle missing values in evaluation data:
`"omit"` drops incomplete rows, `"fail"` errors if any NA is present.}

\item{verbose}{Logical. If TRUE and `na_action="omit"`, prints a message when rows are removed.}
}
\value{
A tibble with evaluation results, containing at least `metric`, `value`, and possibly `time`.
}
\description{
Computes evaluation metrics for a fitted `survdnn` model at one or more time points.
Supported metrics include the concordance index (`"cindex"`), Brier score (`"brier"`),
and integrated Brier score (`"ibs"`).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survdnn.R
\name{build_dnn}
\alias{build_dnn}
\title{Build a Deep Neural Network for Survival Analysis}
\usage{
build_dnn(
  input_dim,
  hidden,
  activation = "relu",
  output_dim = 1L,
  dropout = 0.3,
  batch_norm = TRUE
)
}
\arguments{
\item{input_dim}{Integer. Number of input features.}

\item{hidden}{Integer vector. Sizes of the hidden layers (e.g., c(32, 16)).}

\item{activation}{Character. Name of the activation function to use in each layer.
Supported options: `"relu"`, `"leaky_relu"`, `"tanh"`, `"sigmoid"`, `"gelu"`, `"elu"`, `"softplus"`.}

\item{output_dim}{Integer. Output layer dimension (default = 1).}

\item{dropout}{Numeric between 0 and 1. Dropout rate after each hidden layer
(default = 0.3). Set to 0 to disable dropout.}

\item{batch_norm}{Logical; whether to add `nn_batch_norm1d()` after each
hidden linear layer (default = TRUE).}
}
\value{
A `nn_sequential` object representing the network.
}
\description{
Constructs a multilayer perceptron (MLP) with optional batch normalization
and dropout. Used internally by [survdnn()] to define the model architecture.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.survdnn.R
\name{print.survdnn}
\alias{print.survdnn}
\title{Print a survdnn Model}
\usage{
\method{print}{survdnn}(x, ...)
}
\arguments{
\item{x}{An object of class `"survdnn"`, returned by [survdnn()].}

\item{...}{Ignored (for future compatibility).}
}
\value{
The model object, invisibly.
}
\description{
Pretty prints a fitted `survdnn` model. Displays the formula,
network architecture, training configuration, and final training loss.
}
\examples{
\donttest{
library(survival)
data(veteran, package = "survival")
mod <- survdnn(Surv(time, status) ~
age + karno + celltype, data = veteran, epochs = 20, verbose = FALSE)
print(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluation.R
\name{cv_survdnn}
\alias{cv_survdnn}
\title{K-Fold Cross-Validation for survdnn Models}
\usage{
cv_survdnn(
  formula,
  data,
  times,
  metrics = c("cindex", "ibs"),
  folds = 5,
  .seed = NULL,
  .device = c("auto", "cpu", "cuda"),
  .threads = NULL,
  na_action = c("omit", "fail"),
  verbose = TRUE,
  ...
)
}
\arguments{
\item{formula}{A survival formula, e.g., `Surv(time, status) ~ x1 + x2`.}

\item{data}{A data frame.}

\item{times}{A numeric vector of evaluation time points.}

\item{metrics}{A character vector: any of `"cindex"`, `"brier"`, `"ibs"`.}

\item{folds}{Integer. Number of folds to use.}

\item{.seed}{Optional. Set random seed for reproducibility.}

\item{.device}{Character string indicating the computation device used when fitting the models
in each fold. One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"` uses CUDA if available,
otherwise falls back to CPU.}

\item{.threads}{Optional positive integer. If provided, sets Torch CPU thread
count before each fold fit via `torch::torch_set_num_threads()`.}

\item{na_action}{Character. How to handle missing values within each fold:
`"omit"` drops incomplete rows; `"fail"` errors if any NA is present.}

\item{verbose}{Logical; whether to print cross-validation progress and propagate
verbose messages to fitting/evaluation in each fold (default: TRUE).}

\item{...}{Additional arguments passed to [survdnn()].}
}
\value{
A tibble containing metric values per fold and (optionally) per time point.
}
\description{
Performs cross-validation for a `survdnn` model using the specified evaluation metrics.
}
\examples{
\donttest{
if (requireNamespace("torch", quietly = TRUE) && torch::torch_is_installed()) {
  veteran <- survival::veteran
  cv_survdnn(
    survival::Surv(time, status) ~ age + karno + celltype,
    data = veteran,
    times = c(30, 90, 180),
    metrics = "ibs",
    folds = 3,
    .seed = 42,
    hidden = c(16, 8),
    epochs = 5
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{ibs_survmat}
\alias{ibs_survmat}
\title{Integrated Brier Score (IBS) from a Survival Probability Matrix}
\usage{
ibs_survmat(object, sp_matrix, times)
}
\arguments{
\item{object}{A `Surv` object with observed time and status.}

\item{sp_matrix}{A data frame or matrix of predicted survival probabilities.
Each column corresponds to a time point in `times`.}

\item{times}{A numeric vector of time points. Must match the columns of `sp_matrix`.}
}
\value{
A single numeric value representing the integrated Brier score.
}
\description{
Computes the Integrated Brier Score (IBS) over a set of evaluation time points,
using trapezoidal integration and IPCW adjustment for right-censoring.
}
\examples{
\donttest{
set.seed(123)
library(survival)
data(veteran, package = "survival")
idx <- sample(nrow(veteran), 0.7 * nrow(veteran))
train <- veteran[idx, ]; test <- veteran[-idx, ]
mod <- survdnn(Surv(time, status) ~
age + karno + celltype, data = train, epochs = 50, verbose = FALSE)
pred <- predict(mod, newdata = test, times = c(30, 90, 180), type = "survival")
y_test <- model.response(model.frame(mod$formula, test))
ibs_survmat(y_test, sp_matrix = pred, times = c(30, 90, 180))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{callback_early_stopping}
\alias{callback_early_stopping}
\title{Early stopping callback for survdnn}
\usage{
callback_early_stopping(
  patience = 10L,
  min_delta = 0,
  mode = c("min", "max"),
  verbose = FALSE
)
}
\arguments{
\item{patience}{Integer. Number of epochs with no improvement before stopping.}

\item{min_delta}{Minimum change to qualify as an improvement (default: 0).}

\item{mode}{Character. "min" (for losses) or "max" (for metrics to maximize).}

\item{verbose}{Logical. Whether to print a message when early stopping is triggered.}
}
\value{
A function of the form `function(epoch, current)` that returns TRUE
  if training should stop, FALSE otherwise.
}
\description{
Simple early stopping on a monitored scalar, typically the training loss.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.survdnn.R
\name{summary.survdnn}
\alias{summary.survdnn}
\title{Summarize a Deep Survival Neural Network Model}
\usage{
\method{summary}{survdnn}(object, ...)
}
\arguments{
\item{object}{An object of class `"survdnn"` returned by the [survdnn()] function.}

\item{...}{Currently ignored (for future compatibility).}
}
\value{
Invisibly returns an object of class `"summary.survdnn"`.
}
\description{
Provides a structured summary of a fitted `survdnn` model, including the network architecture,
training configuration, and data characteristics. The summary is printed automatically with
a styled header and sectioned output using \{cli\} and base formatting. The object is returned invisibly.
}
\examples{
\donttest{
set.seed(42)
sim_data <- data.frame(
  age = rnorm(100, 60, 10),
  sex = factor(sample(c("male", "female"), 100, TRUE)),
  trt = factor(sample(c("A", "B"), 100, TRUE)),
  time = rexp(100, 0.05),
  status = rbinom(100, 1, 0.7)
)
mod <- survdnn(Surv(time, status) ~ age + sex + trt, data = sim_data, epochs = 50, verbose = FALSE)
summary(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_survdnn.R
\name{tune_survdnn}
\alias{tune_survdnn}
\title{Tune Hyperparameters for a survdnn Model via Cross-Validation}
\usage{
tune_survdnn(
  formula,
  data,
  times,
  metrics = "cindex",
  param_grid,
  folds = 3,
  .seed = 42,
  .device = c("auto", "cpu", "cuda"),
  na_action = c("omit", "fail"),
  refit = FALSE,
  return = c("all", "summary", "best_model")
)
}
\arguments{
\item{formula}{A survival formula, e.g., `Surv(time, status) ~ x1 + x2`.}

\item{data}{A data frame.}

\item{times}{A numeric vector of evaluation time points.}

\item{metrics}{A character vector of evaluation metrics: "cindex", "brier", or "ibs".
Only the first metric is used for model selection.}

\item{param_grid}{A named list defining hyperparameter combinations to evaluate.
Required names: `hidden`, `lr`, `activation`, `epochs`, `loss`.}

\item{folds}{Number of cross-validation folds (default: 3).}

\item{.seed}{Optional seed for reproducibility.}

\item{.device}{Character string indicating the computation device used when fitting models
during cross-validation and refitting. One of `"auto"`, `"cpu"`, or `"cuda"`. `"auto"`
uses CUDA if available, otherwise falls back to CPU.}

\item{na_action}{Character. How to handle missing values:
`"omit"` drops incomplete rows; `"fail"` errors if any NA is present.}

\item{refit}{Logical. If TRUE, refits the best model on the full dataset.}

\item{return}{One of "all", "summary", or "best_model":
\describe{
  \item{"all"}{Returns the full cross-validation result across all combinations.}
  \item{"summary"}{Returns averaged results per configuration.}
  \item{"best_model"}{Returns the refitted model or best hyperparameters.}
}}
}
\value{
A tibble or model object depending on the `return` value.
}
\description{
Performs k-fold cross-validation over a user-defined hyperparameter grid
and selects the best configuration according to the specified evaluation metric.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_loss.R
\name{plot_loss}
\alias{plot_loss}
\title{Plot Training Loss for a survdnn Model}
\usage{
plot_loss(object, smooth = FALSE, log_y = FALSE, ...)
}
\arguments{
\item{object}{A fitted `survdnn` model.}

\item{smooth}{Logical; if `TRUE`, overlays a smoothed loess curve.}

\item{log_y}{Logical; if `TRUE`, uses a log10 y-scale.}

\item{...}{Reserved for future use.}
}
\value{
A `ggplot` object.
}
\description{
Visualize the evolution of the training loss across epochs for a fitted
`survdnn` model. Helps inspect convergence, instability, or callback effects
(e.g., early stopping).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.survdnn.R
\name{predict.survdnn}
\alias{predict.survdnn}
\title{Predict from a survdnn Model}
\usage{
\method{predict}{survdnn}(object, newdata, times = NULL, type = c("survival", "lp", "risk"), ...)
}
\arguments{
\item{object}{An object of class `"survdnn"` returned by [survdnn()].}

\item{newdata}{A data frame of new observations to predict on.}

\item{times}{Numeric vector of time points at which to compute survival or risk probabilities.
Required if `type = "survival"` or `type = "risk"` for Cox/AFT models.
For CoxTime, `times = NULL` is allowed when `type="survival"` and defaults to event times.}

\item{type}{Character string specifying the type of prediction to return:
\describe{
  \item{"lp"}{Linear predictor. For `"cox"`/`"cox_l2"` this is a log-risk score
  (higher implies worse prognosis, consistent with training sign convention). For `"aft"`,
  this is the predicted location parameter \eqn{\mu(x)} on the log-time scale. For `"coxtime"`,
  this is \eqn{g(t_0, x)} evaluated at a reference time \eqn{t_0} (the first event time).}
  \item{"survival"}{Predicted survival probabilities at each value of `times`.}
  \item{"risk"}{Cumulative risk (1 - survival) at **a single** time point.}
}}

\item{...}{Currently ignored (for future extensions).}
}
\value{
A numeric vector (if `type = "lp"` or `"risk"`), or a data frame
  (if `type = "survival"`) with one row per observation and one column per `times`.
}
\description{
Generate predictions from a fitted `survdnn` model for new data. Supports linear predictors,
survival probabilities at specified time points, or cumulative risk estimates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{cindex_survmat}
\alias{cindex_survmat}
\title{Concordance Index from a Survival Probability Matrix}
\usage{
cindex_survmat(object, predicted, t_star = NULL)
}
\arguments{
\item{object}{A `Surv` object representing the observed survival data.}

\item{predicted}{A data frame or matrix of predicted survival probabilities.
Each column corresponds to a time point (e.g., `t=90`, `t=180`).}

\item{t_star}{A numeric time point corresponding to one of the columns in `predicted`.
If `NULL`, the last column is used.}
}
\value{
A single numeric value representing the C-index.
}
\description{
Computes the time-dependent concordance index (C-index) from a predicted survival matrix
at a fixed time point. The risk is computed as `1 - S(t_star)`.
}
\examples{
\donttest{
library(survival)
data(veteran, package = "survival")
mod <- survdnn(Surv(time, status) ~
age + karno + celltype, data = veteran, epochs = 50, verbose = FALSE)
pred <- predict(mod, newdata = veteran, type = "survival", times = c(30, 90, 180))
y <- model.response(model.frame(mod$formula, veteran))
cindex_survmat(y, pred, t_star = 180)
}
}
